from ecoinvent_interface.release import *
from ecoinvent_interface.core import *

import pyprind


class ABEcoinventRelease(EcoinventRelease):

    def _streaming_download(
            self,
            url: str,
            params: dict,
            directory: Path,
            filename: str,
            headers: Optional[dict] = {},
            zipped: Optional[bool] = False,
    ) -> None:
        """
        Reimplemented _streaming_download with Pyprind.progbar as download bar
        """
        out_filepath = directory / (filename + ".gz" if zipped else filename)
        with requests.get(
                url, stream=True, headers=headers, params=params, timeout=60
        ) as response, open(out_filepath, "wb") as out_file:
            if response.status_code != 200:
                raise requests.exceptions.HTTPError(
                    f"URL '{url}'' returns status code {response.status_code}."
                )
            download = response.raw
            chunk = 128 * 1024

            size = int(response.headers["Content-Length"])
            dl_bar = pyprind.ProgBar(size, title="Downloading from ecoinvent")

            while True:
                segment = download.read(chunk)
                if not segment:
                    break
                dl_bar.update(chunk)
                out_file.write(segment)

        message = """Downloaded file with `_streaming_download`.
    Filename: {filename}
    Directory: {self.storage.dir}
    File size (bytes): {actual}
    Class: {self.__class__.__name__}
    Instance ID: {id(self)}
    Version: {__version__}
    User: {self.username}
        """
        logger.debug(message)

        logger.info("Unzipping download")

        if zipped:
            with open(out_filepath, "rb") as source, open(
                    directory / filename, "w", encoding="utf-8"
            ) as target:
                gzip_fd = gzip.GzipFile(fileobj=source)
                target.write(gzip_fd.read().decode("utf-8-sig"))
            try:
                out_filepath.unlink()
            except PermissionError:
                # Error on Windows during testing
                message = """"Can't automatically delete {out_filepath}
    Please delete manually"""
                warnings.warn(message)

    def get_release(
        self,
        version: str,
        system_model: str,
        release_type: ReleaseType,
        extract: Optional[bool] = True,
        force_redownload: Optional[bool] = False,
        fix_version: Optional[bool] = True,
    ) -> Path:
        if not isinstance(release_type, ReleaseType):
            raise ValueError("`release_type` must be an instance of `ReleaseType`")

        abbr = SYSTEM_MODELS.get(system_model, system_model)
        filename = release_type.filename(version=version, system_model_abbr=abbr)
        available_files = self._filename_dict(version=version)

        if filename not in available_files:
            # Sometimes the filename prediction doesn't work, as not every filename
            # follows our patterns. But these exceptions are unpredictable, it's
            # just easier to find the closest match and log the correction
            # than build a catalogue of exceptions.
            possible = sorted(
                [
                    (damerau_levenshtein(filename, maybe), maybe)
                    for maybe in available_files
                ]
            )[0]
            if possible[0] <= 3:
                logger.info(
                    f"Using close match {possible[1]} for predicted filename {filename}"
                )
                filename = possible[1]
            else:
                ERROR = """"Can't find predicted filename {filename}.
    Closest match is {possible[1]}.
    Filenames for this version:""" + "\n\t".join(
                    available_files
                )
                raise ValueError(ERROR)

        cached = filename in self.storage.catalogue
        result_path = self._download_and_cache(
            filename=filename,
            uuid=available_files[filename]["uuid"],
            modified=available_files[filename]["modified"],
            expected_size=available_files[filename]["size"],
            url_namespace="r",
            extract=extract,
            force_redownload=force_redownload,
            version=version,
            system_model=system_model,
            kind="release",
        )

        SPOLD_FILES = (ReleaseType.ecospold, ReleaseType.lci, ReleaseType.lcia)
        if fix_version and release_type in SPOLD_FILES and not cached:
            major, minor = major_minor_from_string(version)
            if (result_path / "datasets").is_dir():
                logger.info("Fixing versions in unit process datasets")

                for filepath in pyprind.prog_bar(list((result_path / "datasets").iterdir()),
                                                 title="Fixing versions in unit process data"):
                    if not filepath.suffix.lower() == ".spold":
                        continue
                    fix_version_upr(
                        filepath=filepath, major_version=major, minor_version=minor
                    )
            if (result_path / "MasterData").is_dir():
                logger.info("Fixing versions in master data")
                for filepath in pyprind.prog_bar(list((result_path / "MasterData").iterdir()),
                                                 title="Fixing versions in master data"):
                    if not filepath.suffix.lower() == ".xml":
                        continue
                    fix_version_meta(
                        filepath=filepath, major_version=major, minor_version=minor
                    )

        return result_path



from qtpy import QtWidgets, QtCore
from qtpy.QtCore import Qt

import bw2data as bd
import pandas as pd

from activity_browser import signals, actions
from activity_browser.ui import widgets, core, delegates


class ImpactCategoriesPane(widgets.ABAbstractPane):
    title = "Impact Categories"
    unique = True

    def __init__(self, parent=None):
        super().__init__(parent)
        self.view = ImpactCategoriesView()
        self.model = ImpactCategoriesModel()
        self.view.setModel(self.model)

        self.view.setSelectionMode(QtWidgets.QTableView.SingleSelection)
        self.view.setDragEnabled(True)
        self.view.setDragDropMode(QtWidgets.QTableView.DragDropMode.DragOnly)
        self.view.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)

        self.search = widgets.ABLineEdit(self)
        self.search.setMaximumHeight(30)
        self.search.setPlaceholderText("Quick Search")

        self.search.textChangedDebounce.connect(self.view.setAllFilter)

        self.build_layout()
        self.connect_signals()
        self.load()

    def build_layout(self):
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.search)
        layout.addWidget(self.view)
        layout.setContentsMargins(5, 0, 5, 5)

        self.setLayout(layout)

    def connect_signals(self):
        signals.meta.methods_changed.connect(self.sync)
        signals.project.changed.connect(self.sync)
        signals.database_read_only_changed.connect(self.sync)

    def load(self):
        self.model.setDataFrame(self.build_df())
        self.model.group(1)
        self.view.setColumnHidden(1, True)
        self.view.setColumnHidden(2, True)
        self.view.setColumnHidden(3, True)
        self.view.sortByColumn(1, Qt.SortOrder.AscendingOrder)

    def sync(self):
        self.model.setDataFrame(self.build_df())

    def build_df(self):
        df = pd.DataFrame(bd.methods.values())
        df["_method_name"] = bd.methods.keys()

        df["name"] = df["_method_name"].apply(lambda x: x[-1])
        df["groups"] = df["_method_name"].apply(lambda x: x[:-1])

        cols = ["name", "groups", "unit", "num_cfs", "_method_name"]

        if df.empty:
            return pd.DataFrame(columns=cols)

        return df[cols]


class ImpactCategoriesView(widgets.ABTreeView):
    defaultColumnDelegates = {
        "groups": delegates.ListDelegate,
    }

    class ContextMenu(widgets.ABMenu):
        menuSetup = [
            lambda m, p: m.add(actions.MethodNew),
            lambda m: m.addSeparator(),
            lambda m, p: m.add(actions.MethodOpen, p.selected_impact_categories,
                               text="Open impact category" if len(p.selected_impact_categories) == 1 else "Open impact categories",
                               enable=len(p.selected_impact_categories) > 0
                               ),
            lambda m, p: m.add(actions.MethodDelete, p.selected_impact_categories,
                               text="Delete impact category" if len(
                                   p.selected_impact_categories) == 1 else "Delete impact categories",
                               enable=len(p.selected_impact_categories) > 0
                               ),
            lambda m, p: m.add(actions.MethodDuplicate, p.selected_impact_categories,
                               text="Duplicate impact category",
                               enable=len(p.selected_impact_categories) == 1
                               ),
            lambda m, p: m.add(actions.MethodRename, p.selected_impact_categories,
                               text="Rename impact category",
                               enable=len(p.selected_impact_categories) == 1
                               ),
        ]

        @staticmethod
        def get_functional_unit_amount(key):
            from activity_browser.bwutils import refresh_node
            excs = list(refresh_node(key).upstream(["production"]))
            exc = excs[0] if len(excs) == 1 else {}
            return exc.get("amount", 1.0)

        @property
        def database_name(self):
            return self.parent().parent().database.name

    @property
    def selected_impact_categories(self):
        indices = [i for i in self.selectedIndexes() if i.column() == 0]
        impact_categories = []

        for index in indices:
            impact_categories.extend(self.model().get_impact_categories(index))

        return list(set(impact_categories))

    def mouseDoubleClickEvent(self, event) -> None:
        if self.selected_impact_categories:
            actions.MethodOpen.run(self.selected_impact_categories)


class ImpactCategoriesItem(widgets.ABDataItem):
    def flags(self, col: int, key: str):
        """
        Returns the item flags for the given column and key.

        Args:
            col (int): The column index.
            key (str): The key for which to return the flags.

        Returns:
            QtCore.Qt.ItemFlags: The item flags.
        """
        return super().flags(col, key) | Qt.ItemFlag.ItemIsDragEnabled


class ImpactCategoriesBranchItem(widgets.ABBranchItem):
    def flags(self, col: int, key: str):
        """
        Returns the item flags for the given column and key.

        Args:
            col (int): The column index.
            key (str): The key for which to return the flags.

        Returns:
            QtCore.Qt.ItemFlags: The item flags.
        """
        return super().flags(col, key) | Qt.ItemFlag.ItemIsDragEnabled


class ImpactCategoriesModel(widgets.ABItemModel):
    dataItemClass = ImpactCategoriesItem
    branchItemClass = ImpactCategoriesBranchItem

    def mimeData(self, indices: list[QtCore.QModelIndex]):
        """
        Returns the mime data for the given indices.

        Args:
            indices (list[QtCore.QModelIndex]): The indices to get the mime data for.

        Returns:
            core.ABMimeData: The mime data.
        """
        data = core.ABMimeData()
        names = []

        for index in indices:
            names += self.get_impact_categories(index)

        data.setPickleData("application/bw-methodnamelist", list(set(names)))
        return data

    def get_impact_categories(self, index: QtCore.QModelIndex):
        if isinstance(index.internalPointer(), self.dataItemClass):
            return [index.internalPointer()["_method_name"]]

        ics = []
        for i, child in enumerate(index.internalPointer().children().values()):
            child_index = self.createIndex(i, 0, child)
            ics += self.get_impact_categories(child_index)
        return ics




from uuid import uuid4

from qtpy import QtWidgets

import bw2data as bd

from bw_functional import Process

from activity_browser import application, bwutils
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.ui.icons import qicons


class ActivityNewProduct(ABAction):
    """
    ABAction to create a new product for an activity.

    This action prompts the user to supply a name, unit, and location for the new product.
    If the user cancels or does not provide a name, the action is aborted.
    Otherwise, it creates a new product associated with the given activity.

    Attributes:
        icon (QIcon): The icon representing this action.
        text (str): The display text for this action.
    """

    icon = qicons.add
    text = "Create product"

    @staticmethod
    @exception_dialogs
    def run(activities: list[tuple | int | bd.Node], product_type: str = "product"):
        """
        Execute the action to create a new product.

        This method iterates over the provided activities, ensuring each is a `Process`.
        It prompts the user to input details for the new product. If valid details are provided,
        a new product is created and saved.

        Args:
            activities (list[tuple | int | bd.Node]): A list of activities to process.
            product_type (str, optional): The type of the new product. Defaults to "product".

        Raises:
            AssertionError: If an activity is not of type `Process`.
        """
        activities = [bwutils.refresh_node(activity) for activity in activities]

        for act in activities:
            assert isinstance(act, Process), "Cannot create new product for non-process type"
            # Ask the user to provide a name for the new product
            dialog = NewProductDialog(act, product_type, application.main_window)
            # If the user cancels, skip to the next activity
            if dialog.exec_() != QtWidgets.QDialog.Accepted:
                continue
            name, unit, location = dialog.get_new_process_data()
            # If no name is provided, skip to the next activity
            if not name:
                continue

            # Create the new product
            new_prod_data = {
                "product": name,
                "unit": unit,
                "location": location,
                "type": product_type,
            }
            new_product = act.new_product(code=uuid4().hex, **new_prod_data)
            new_product.save()


class NewProductDialog(QtWidgets.QDialog):
    """
    A dialog for gathering parameters to create a new product.

    This dialog allows the user to input the product name, unit, and location.
    It validates the input and provides options to either create the product or cancel the operation.
    """

    def __init__(self, activity: bd.Node, product_type: str, parent: QtWidgets.QWidget = None):
        """
        Initialize the NewProductDialog.

        Args:
            activity (bd.Node): The activity for which the product is being created.
                             Used to prefill the location field and set the dialog title.
            product_type (str): The type of the new product ("product", "waste").
            parent (QtWidgets.QWidget, optional): The parent widget for the dialog. Defaults to None.
        """
        super().__init__(parent)

        # Set the dialog window title
        self.setWindowTitle(f"Create {product_type} for {activity['name']}")

        # Input fields for product details
        self.name_edit = QtWidgets.QLineEdit()
        self.unit_edit = QtWidgets.QLineEdit("kilogram")  # Default unit is "kilogram"
        self.location_edit = QtWidgets.QLineEdit(activity.get("location", ""))  # Prefill location if available

        # Buttons for user actions
        self.ok_button = QtWidgets.QPushButton("Create")
        self.ok_button.clicked.connect(self.accept)  # Connect the "Create" button to accept the dialog
        self.ok_button.setEnabled(False)  # Initially disable the button until a name is entered

        self.cancel_button = QtWidgets.QPushButton("Cancel")
        self.cancel_button.clicked.connect(self.reject)  # Connect the "Cancel" button to reject the dialog

        # Set up signals and layout
        self.connect_signals()
        self.build_layout()

    def connect_signals(self):
        """
        Connect signals to their respective handlers.

        - Enables the "Create" button only when the name field is not empty.
        """
        self.name_edit.textChanged.connect(lambda x: self.ok_button.setEnabled(bool(x)))

    def build_layout(self):
        """
        Build and set the layout for the dialog.

        The layout includes labels and input fields for product name, unit, and location,
        as well as "Create" and "Cancel" buttons.
        """
        layout = QtWidgets.QGridLayout()
        layout.addWidget(QtWidgets.QLabel("Name"), 0, 0)
        layout.addWidget(self.name_edit, 0, 1)

        layout.addWidget(QtWidgets.QLabel("Unit"), 1, 0)
        layout.addWidget(self.unit_edit, 1, 1)

        layout.addWidget(QtWidgets.QLabel("Location"), 2, 0)
        layout.addWidget(self.location_edit, 2, 1)

        layout.addWidget(self.ok_button, 3, 0)
        layout.addWidget(self.cancel_button, 3, 1)

        self.setLayout(layout)

    def get_new_process_data(self) -> tuple[str, str, str]:
        """
        Retrieve the parameters entered by the user.

        Returns:
            tuple[str, str, str]: A tuple containing the product name, unit, and location.
        """
        return (
            self.name_edit.text(),
            self.unit_edit.text(),
            self.location_edit.text()
        )



from typing import List
from logging import getLogger

from qtpy import QtWidgets

from activity_browser import application
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.mod import bw2data as bd
from activity_browser.ui.icons import qicons

log = getLogger(__name__)


class MethodMetaModify(ABAction):
    """
    """

    icon = qicons.delete
    text = "Modify Impact Category metadata"

    @staticmethod
    @exception_dialogs
    def run(method_name: tuple[str], key: str, value: str):
        if method_name not in bd.methods:
            log.warning(f"Can't modify metadata for method {method_name} - method not found")
            return

        bd.methods[method_name][key] = value
        bd.methods.flush()



from logging import getLogger

from activity_browser import bwutils
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.ui.icons import qicons

from bw_functional import Process
from bw2data import databases

log = getLogger(__name__)


class ProcessPropertyRemove(ABAction):
    """
    Remove a specified property from a process and its associated products.

    This method refreshes the given process, validates its type, and checks if the specified
    property exists. If the property is an allocation property, it resets the allocation to
    the database's default allocation. The property is then removed from all products of the process.

    Args:
        process (tuple | int | Process): The process from which the property will be removed.
            Can be a tuple (key), integer (id), or Process object.
        property_name (str): The name of the property to remove.

    Raises:
        ValueError: If the provided process is not of type Process.

    Logs:
        Warning: If the specified property is not found in the process.
    """

    icon = qicons.delete
    text = "Remove property"

    @staticmethod
    @exception_dialogs
    def run(process: tuple | int | Process, property_name: str):
        process = bwutils.refresh_node(process)
        if not isinstance(process, Process):
            raise ValueError(f"Expected a Process-type activity, got {type(process)} instead")

        allocate = property_name == process.get("allocation")

        if property_name not in process.available_properties():
            log.warning(f"Property '{property_name}' not found in process {process.key}.")
            return

        if allocate:
            process["allocation"] = databases[process["database"]].get("default_allocation", "equal")
            process.save()

        # Remove the property from all products of the process
        for product in process.products():
            if property_name not in product.get("properties", {}):
                continue

            del product["properties"][property_name]
            product.save()




from logging import getLogger

from activity_browser import application
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.ui.icons import qicons
from activity_browser.ui.application import global_shortcut

log = getLogger(__name__)


class MetaDataStoreOpen(ABAction):

    icon = qicons.right
    text = "Open activity / activities"

    @staticmethod
    @global_shortcut("Ctrl+Shift+M")
    @exception_dialogs
    def run():
        from activity_browser.layouts import pages
        page = pages.MetaDataStorePage()
        central = application.main_window.centralWidget()
        central.addToGroup("DEBUG", page)



# -*- coding: utf-8 -*-
from qtpy import QtCore, QtWidgets


class CheckboxDelegate(QtWidgets.QStyledItemDelegate):
    def __init__(self, parent=None):
        super().__init__(parent)

    def createEditor(self, parent, option, index):
        return None

    def paint(self, painter, option, index):
        """Paint the cell with a styled option button, showing a checkbox

        See links below for inspiration:
        https://stackoverflow.com/a/11778012
        https://stackoverflow.com/q/15235273

        NOTE: PyQt 5.9.2 needs to treat OSX different from others.
         qtpy 5.13.1 and higher no longer has this issue.
        """
        painter.save()
        value = bool(index.data(QtCore.Qt.DisplayRole))
        button = QtWidgets.QStyleOptionButton()
        button.state = QtWidgets.QStyle.State_Enabled
        button.state |= (
            QtWidgets.QStyle.State_Off if not value else QtWidgets.QStyle.State_On
        )
        painter.translate(QtCore.QPoint(option.rect.left(), option.rect.center().y()))
        style = option.widget.style()
        style.drawControl(QtWidgets.QStyle.CE_CheckBox, button, painter, option.widget)
        painter.restore()



from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.mod import bw2data as bd
from activity_browser.ui.icons import qicons


class CFAmountModify(ABAction):
    """
    ABAction to modify the amount of a characterization factor within a method. Updates the CF-Tuple's second value
    directly if there's no uncertainty dict. Otherwise, changes the "amount" from the uncertainty dict.
    """

    icon = qicons.edit
    text = "Modify amount"

    @staticmethod
    @exception_dialogs
    def run(method: tuple | bd.Method, key: int | tuple, amount: float):
        if isinstance(method, bd.Method):
            method = method.name

        method = bd.Method(method)
        method_dict = {cf[0]: cf[1] for cf in method.load()}

        if isinstance(method_dict[key], dict):
            method_dict[key]["amount"] = amount
        else:
            method_dict[key] = amount

        method.write(list(method_dict.items()))



from activity_browser import application
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.ui.icons import qicons


class DatabaseExplorerOpen(ABAction):
    """
    ABAction to delete a database from the project. Asks the user for confirmation. If confirmed, instructs the
    DatabaseController to delete the database in question.
    """

    icon = qicons.delete
    text = "Explore database"
    tool_tip = "Delete this database from the project"

    @staticmethod
    @exception_dialogs
    def run(db_name: str):
        from activity_browser.layouts.panes import DatabaseExplorerPane
        db_explorer = DatabaseExplorerPane(db_name, application.main_window)
        db_explorer.show()



from qtpy import QtWidgets, QtCore

from activity_browser import application, bwutils
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.ui.icons import qicons

from bw_functional import Process


class ProcessPropertyModify(ABAction):
    """
    Modify a property for all the products of a process.

    This method refreshes the given process, validates its type, and opens a dialog
    for the user to modify a property. If the property already exists, the dialog
    is pre-populated with its current values. The updated property is then applied
    to all products of the process.

    Args:
        process (tuple | int | Process): The process to modify. Can be a tuple, integer, or Process object.
        property_name (str, optional): The name of the property to modify. Defaults to None.

    Raises:
        ValueError: If the provided process is not of type Process.
    """

    icon = qicons.edit
    text = "Modify property"

    @staticmethod
    @exception_dialogs
    def run(process: tuple | int | Process,
            property_name: str = None
            ):

        process = bwutils.refresh_node(process)
        if not isinstance(process, Process):
            raise ValueError(f"Expected a Process-type activity, got {type(process)} instead")

        prop_dialog = PropertyDialog(process)

        # if the property already exists, populate the dialog with the existing values
        if property_name in process.available_properties():
            prop = process.property_template(property_name)
            prop_dialog.prop_name.setText(property_name)
            prop_dialog.prop_unit.setText(prop["unit"])
            prop_dialog.normalize_check.setChecked(prop.get("normalize", True))

        # show the dialog to the user
        if prop_dialog.exec_() != QtWidgets.QDialog.DialogCode.Accepted:
            return

        name_changed = prop_dialog.name != property_name if property_name else False

        for product in process.products():
            # make sure the dictionaries are in place
            product["properties"] = product.get("properties", {})
            product["properties"][prop_dialog.name] = product["properties"].get(property_name, {})

            prop = {
                "unit": prop_dialog.prop["unit"],
                "normalize": prop_dialog.prop["normalize"],
                "amount": product["properties"][prop_dialog.name].get("amount", 1.0),
            }

            # update the property with the new values
            product["properties"][prop_dialog.name] = prop

            # if the name has changed, remove the old property
            if name_changed and property_name in product["properties"]:
                del product["properties"][property_name]

            product.save()


class PropertyDialog(QtWidgets.QDialog):
    name: str | None = None
    prop: dict | None = None

    def __init__(self, process: Process):
        super().__init__(application.main_window)
        self.process = process

        self.setWindowTitle("Add Property")

        self.prop_name = QtWidgets.QLineEdit(self)
        self.prop_name.setPlaceholderText("Property name")
        self.prop_name.textChanged.connect(self.validate)

        self.prop_unit = QtWidgets.QLineEdit(self)
        self.prop_unit.setPlaceholderText("Property unit")
        self.prop_unit.textChanged.connect(self.validate)

        self.normalize_label = QtWidgets.QLabel(" / amount", self)
        self.normalize_label.setVisible(False)
        self.normalize_check = QtWidgets.QCheckBox("per amount")
        self.normalize_check.setChecked(True)
        self.normalize_check.toggled.connect(self.normalize_label.setVisible)
        self.normalize_check.toggled.connect(self.validate)

        self.buttons = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel,
        )
        self.buttons.button(QtWidgets.QDialogButtonBox.Ok).setEnabled(False)
        self.buttons.accepted.connect(self.accept)
        self.buttons.rejected.connect(self.reject)

        h_layout = QtWidgets.QHBoxLayout()
        h_layout.addWidget(self.prop_unit)
        h_layout.addWidget(self.normalize_label)

        self.layout = QtWidgets.QVBoxLayout()
        self.layout.addWidget(self.prop_name)
        self.layout.addLayout(h_layout)
        self.layout.addWidget(self.normalize_check, alignment=QtCore.Qt.AlignmentFlag.AlignRight)
        self.layout.addWidget(self.buttons)

        self.setLayout(self.layout)

    def validate(self):
        if (
            self.prop_name.text() and
            self.prop_unit.text() and
            self.prop_name.text() not in self.process.get("properties", [])
        ):
            self.name = self.prop_name.text()
            self.prop = {
                "unit": self.prop_unit.text(),
                "normalize": self.normalize_check.isChecked(),
            }
            self.buttons.button(QtWidgets.QDialogButtonBox.Ok).setEnabled(True)
        else:
            self.name = None
            self.prop = None
            self.buttons.button(QtWidgets.QDialogButtonBox.Ok).setEnabled(False)



from logging import getLogger
from typing import List

from qtpy import QtWidgets

from activity_browser import application
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.ui import widgets
from activity_browser.bwutils import exporters
from activity_browser.ui.core import threading

log = getLogger(__name__)


class DatabaseExportBW2Package(ABAction):
    """
    ABAction to export database(s) to BW2Package format (.bw2package).
    """

    # icon = icons.qicons.export_db
    text = "Export to .bw2package"
    tool_tip = "Export database(s) to BW2Package format"

    @classmethod
    @exception_dialogs
    def run(cls, db_names: List[str] = None):
        if db_names is None:
            import bw2data as bd
            dialog = widgets.ABDatabaseSelectionDialog(
                parent=application.main_window,
                databases=sorted(bd.databases),
                title="Select databases to export to BW2Package"
            )
            if dialog.exec_() == QtWidgets.QDialog.Accepted:
                db_names = dialog.get_selected_databases()
            else:
                return

        # Get export directory or file from user
        if len(db_names) == 1:
            # Single database - suggest a filename
            suggested_name = f"{db_names[0]}.bw2package"
            path, _ = QtWidgets.QFileDialog.getSaveFileName(
                parent=application.main_window,
                caption=f'Export database "{db_names[0]}" to BW2Package',
                directory=suggested_name,
                filter='Brightway2 Database Package (*.bw2package);; All files (*.*)'
            )
        else:
            # Multiple databases - ask for directory
            path = QtWidgets.QFileDialog.getExistingDirectory(
                parent=application.main_window,
                caption=f'Select directory to export {len(db_names)} databases',
            )
        
        if not path:
            return

        # Show export dialog
        context = {
            "db_names": db_names,
            "path": path,
        }
        export_dialog = ExportBW2PackageSetup(
            parent=application.main_window,
            title="Export to BW2Package",
            context=context
        )
        export_dialog.show()


class ExportBW2PackageSetup(widgets.ABWizard):
    """Wizard for exporting databases to BW2Package format."""

    class ExportPage(widgets.ABThreadedWizardPage):
        """Wizard page to export the selected database(s) to BW2Package."""
        title = "Exporting Database(s)"
        subtitle = "Exporting database(s) to .bw2package file(s)"

        class Thread(threading.ABThread):
            """Thread to handle the export process."""
            
            def run_safely(self, db_names: List[str], path: str):
                """Export the database(s) to BW2Package."""
                for db_name in db_names:
                    try:
                        success = exporters.store_database_as_package(db_name, path)
                        if success:
                            log.info(f"Successfully exported database '{db_name}' to BW2Package")
                        else:
                            log.error(f"Failed to export database '{db_name}'")
                            raise RuntimeError(f"Database '{db_name}' not found")
                    except Exception as e:
                        log.error(f"Failed to export database '{db_name}': {e}")
                        raise

        def initializePage(self, context: dict):
            """Start the export thread."""
            self.thread.start(context["db_names"], context["path"])

    pages = [ExportPage]



from .activity_details import ActivityDetailsPage
from .welcome import WelcomePage
from .calculation_setup import CalculationSetupPage
from .impact_category_details import ImpactCategoryDetailsPage
from .lca_results import LCAResultsPage
from .parameters import ParametersPage
from .metadatastore import MetaDataStorePage



# -*- coding: utf-8 -*-
from .dataframe import (scenario_names_from_df, scenario_replace_databases,
                        superstructure_from_arrays, superstructure_from_scenario_exchanges)
from .excel import get_sheet_names, import_from_excel
from .file_dialogs import ABPopup
from .file_imports import ABCSVImporter, ABFeatherImporter, ABFileImporter
from .manager import SuperstructureManager
from .mlca import SuperstructureContributions, SuperstructureMLCA
from .utils import SUPERSTRUCTURE, _time_it_, edit_superstructure_for_string, parameters_to_sdf



from logging import getLogger
from time import time

import pandas as pd
from qtpy import QtWidgets, QtCore, QtGui
from qtpy.QtCore import Qt

import bw2data as bd

from activity_browser import actions, ui, signals, application
from activity_browser.settings import project_settings
from activity_browser.ui import core, widgets, delegates
from activity_browser.bwutils import AB_metadata, database_is_locked, database_is_legacy

log = getLogger(__name__)

DEFAULT_STATE = {
    "columns": ["activity", "product", "Type", "Unit", "Location"],
    "visible_columns": ["activity", "product", "type", "unit", "location"],
}

NODETYPES = {
    "all_nodes": [],
    "processes": ["process", "multifunctional", "processwithreferenceproduct", "nonfunctional"],
    "products": ["product", "processwithreferenceproduct", "waste"],
    "biosphere": ["natural resource", "emission", "inventory indicator", "economic", "social"],
}


class DatabaseProductsPane(widgets.ABAbstractPane):
    """
    A widget that displays products related to a specific database.

    Attributes:
        database (bd.Database): The database to display products for.
        model (ProductModel): The model containing the data for the products.
        table_view (ProductView): The view displaying the products.
        search (widgets.ABLineEdit): The search bar for quick search.
    """
    def __init__(self, parent, db_name: str):
        """
        Initializes the DatabaseProductsPane widget.

        Args:
            parent (QtWidgets.QWidget): The parent widget.
            db_name (str): The name of the database to display products for.
        """
        self.name = "database_products_pane_" + db_name

        super().__init__(parent)
        self.database = bd.Database(db_name)
        self.title = db_name
        self.model = ProductModel(self)

        # Create the QTableView and set the model
        self.table_view = ProductView(self, db_name=db_name)
        self.table_view.setModel(self.model)

        self.search = widgets.ABLineEdit(self)
        self.search.setMaximumHeight(30)
        self.search.setPlaceholderText("Quick Search")

        # Create loading indicator with spinner
        self.loading_spinner = QtWidgets.QProgressBar()
        self.loading_spinner.setRange(0, 0)  # Indeterminate/busy indicator
        self.loading_spinner.setTextVisible(False)
        self.loading_spinner.setMaximumWidth(200)
        self.loading_spinner.setMaximumHeight(20)
        
        self.loading_label = widgets.ABLabel("Loading database...")
        self.loading_label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        font = self.loading_label.font()
        font.setPointSize(14)
        self.loading_label.setFont(font)
        self.loading_label.setStyleSheet("color: gray; padding: 10px;")

        self.build_layout()
        self.connect_signals()
        self.update_loading_state()
        self.sync()

    def build_layout(self):
        # Create a stacked layout to switch between loading and table view
        self.stacked_layout = QtWidgets.QStackedLayout()
        
        # Page 0: Loading indicator with spinner
        loading_widget = QtWidgets.QWidget(self)
        loading_layout = QtWidgets.QVBoxLayout(loading_widget)
        loading_layout.addStretch()
        loading_layout.addWidget(self.loading_spinner, alignment=Qt.AlignmentFlag.AlignCenter)
        loading_layout.addWidget(self.loading_label)
        loading_layout.addStretch()
        self.stacked_layout.addWidget(loading_widget)
        
        # Page 1: Table view
        table_widget = QtWidgets.QWidget(self)
        table_layout = QtWidgets.QVBoxLayout(table_widget)
        table_layout.setSpacing(0)
        table_layout.setContentsMargins(0, 0, 0, 0)
        table_layout.addWidget(self.table_view)
        self.stacked_layout.addWidget(table_widget)

        layout = QtWidgets.QVBoxLayout(self)
        layout.addWidget(self.search)
        layout.addLayout(self.stacked_layout)

        # Set the table view as the central widget of the window
        self.setLayout(layout)

    def connect_signals(self):
        AB_metadata.synced.connect(self.on_metadata_changed)
        signals.database.deleted.connect(self.on_database_deleted)

        self.table_view.filtered.connect(self.search_error)
        self.search.textChangedDebounce.connect(self.table_view.setAllFilter)

    def on_metadata_changed(self, added, updated, deleted):
        # Check if primary data has finished loading
        self.update_loading_state()
        
        if any(db == self.database.name for db, code in added | updated | deleted):
            self.sync()

    def update_loading_state(self):
        """
        Updates the loading state based on whether primary metadata has loaded.
        Shows the loading indicator if primary data is still loading, otherwise shows the table.
        """
        if AB_metadata.loader.secondary_status == "done":
            # Show table view
            self.stacked_layout.setCurrentIndex(1)
        else:
            # Show loading indicator
            self.stacked_layout.setCurrentIndex(0)

    def sync(self):
        """
        Synchronizes the widget with the current state of the database.
        """
        t = time()
        df = self.build_df()
        self.model.setDataFrame(df)
        for col in df.columns:
            index = self.model.columns().index(col)
            if df[col].isna().all():
                self.table_view.hideColumn(index)
            else:
                self.table_view.showColumn(index)

        log.debug(f"Synced DatabaseProductsPane in {time() - t:.2f} seconds")

    def build_df(self) -> pd.DataFrame:
        """
        Builds a DataFrame from the database products.

        Returns:
            pd.DataFrame: The DataFrame containing the products data.
        """
        t = time()
        cols = ["name", "key", "processor", "product", "type", "unit", "location", "id", "categories", "properties"]
        df = AB_metadata.get_database_metadata(self.database.name, cols)

        processors = set(df["processor"].dropna().unique())
        df = df.drop(processors, errors="ignore")

        if not df.properties.isna().all():
            props_df = df[df.properties.notna()]
            props_df = pd.DataFrame(list(props_df.get("properties")), index=props_df.key)
            props_df.rename(lambda col: f"property_{col}", axis="columns", inplace=True)

            df = df.merge(
                props_df,
                left_on="key",
                right_index=True,
                how="left",
            )

        cols = ["name", "product", "categories", "unit", "location", "key", "processor", "type",]
        cols += [col for col in df.columns if col.startswith("property")]

        log.debug(f"Built DatabaseProductsPane dataframe in {time() - t:.2f} seconds")

        return df[cols]

    def on_database_deleted(self, db_name: str):
        """
        Handles the database deleted signal by closing the widget if the database is deleted.

        Args:
            db_name (str): The name of the deleted database.
        """
        if db_name == self.database.name:
            self.deleteLater()

    def event(self, event):
        """
        Handles the event to save the state to settings on deferred delete.

        Args:
            event: The event to handle.

        Returns:
            bool: True if the event was handled, False otherwise.
        """
        if event.type() == QtCore.QEvent.Type.DeferredDelete:
            self.save_state_to_settings()

        return super().event(event)

    def save_state_to_settings(self):
        """
        Saves the state of the table view to the project settings.
        """
        project_settings.settings["database_explorer"] = project_settings.settings.get("database_explorer", {})
        project_settings.settings["database_explorer"][self.database.name] = self.table_view.saveState()
        project_settings.write_settings()

    def get_state_from_settings(self):
        """
        Gets the state from the project settings.

        Returns:
            dict: The state of the table view.
        """
        return DEFAULT_STATE

    def search_error(self, reset=False):
        """
        Handles the search error by changing the search bar color.

        Args:
            reset (bool, optional): Whether to reset the search bar color. Defaults to False.
        """
        if reset:
            self.search.setPalette(application.palette())
            return

        palette = self.search.palette()
        palette.setColor(QtGui.QPalette.ColorRole.Base, QtGui.QColor(255, 128, 128))
        self.search.setPalette(palette)


class ProductView(ui.widgets.ABTreeView):
    """
    A view that displays the products in a tree structure.

    Attributes:
        defaultColumnDelegates (dict): The default column delegates for the view.
    """
    defaultColumnDelegates = {
        "categories": delegates.ListDelegate,
        "key": delegates.StringDelegate,
        "processor": delegates.StringDelegate,
    }

    class ContextMenu(ui.widgets.ABMenu):
        menuSetup = [
            lambda m, p: m.add(actions.ActivityOpen, p.selected_activities,
                               text="Open process" if len(p.selected_activities) == 1 else "Open processes",
                               enable=len(p.selected_activities) > 0
                               ),
            lambda m, p: m.add(actions.ActivityGraph, p.selected_activities,
                               enable=len(p.selected_activities) > 0,
                               ),
            lambda m: m.addSeparator(),
            lambda m, p: m.add(actions.ActivityNewProcess, p.db_name,
                               enable=not database_is_locked(p.db_name),
                               ),
            lambda m, p: m.add(actions.ActivityDuplicate, p.selected_activities,
                               text="Duplicate process" if len(p.selected_activities) == 1 else "Duplicate processes",
                               enable=len(p.selected_activities) > 0 and not database_is_locked(p.db_name),
                               ),
            lambda m, p: m.add(actions.ActivityDuplicateToDB, p.selected_activities,
                               text="Duplicate process to database" if len(p.selected_activities) == 1 else "Duplicate processes to database",
                               enable=len(p.selected_activities) > 0 and not database_is_locked(p.db_name),
                               ),
            lambda m: m.addSeparator(),
            lambda m, p: m.add(actions.ActivityDelete, p.selected_activities,
                               text="Delete process" if len(p.selected_activities) == 1 else "Delete processes",
                               enable=len(p.selected_activities) > 0 and not database_is_locked(p.db_name),
                               ),
            lambda m, p: m.add(actions.ActivityDelete, p.selected_products,
                               text="Delete product" if len(p.selected_products) == 1 else "Delete products",
                               enable=len(p.selected_products) > 0 and not
                               database_is_locked(p.db_name) and not
                               database_is_legacy(p.db_name),
                               ),
            lambda m: m.addSeparator(),
            lambda m, p: m.add(actions.CSNew,
                               functional_units=[{prod: m.get_functional_unit_amount(prod)} for prod in p.selected_products],
                               enable=len(p.selected_products) > 0,
                               text="Create setup"
                               ),
            lambda m, p: m.add(actions.ActivitySDFToClipboard, p.selected_products,
                               enable=len(p.selected_products) > 0,
                               ),
        ]

        @staticmethod
        def get_functional_unit_amount(key):
            from activity_browser.bwutils import refresh_node
            excs = list(refresh_node(key).upstream(["production"]))
            exc = excs[0] if len(excs) == 1 else {}
            return exc.get("amount", 1.0)

    def __init__(self, parent: DatabaseProductsPane, db_name: str):
        """
        Initializes the ProductView.

        Args:
            parent (DatabaseProductsPane): The parent widget.
            db_name (str): The name of the database.
        """
        super().__init__(parent)
        self.setSortingEnabled(True)
        self.setDragEnabled(True)
        self.setDragDropMode(QtWidgets.QTableView.DragDropMode.DragOnly)
        self.setSelectionBehavior(ui.widgets.ABTreeView.SelectionBehavior.SelectRows)
        self.setSelectionMode(ui.widgets.ABTreeView.SelectionMode.ExtendedSelection)

        self.db_name = db_name

        self.propertyDelegate = delegates.PropertyDelegate(self)

    def setDefaultColumnDelegates(self):
        """
        Sets the default column delegates for the view.
        """
        super().setDefaultColumnDelegates()

        columns = self.model().columns()
        for i, col_name in enumerate(columns):
            if not col_name.startswith("property_"):
                continue
            # Set the delegate for property columns
            self.setItemDelegateForColumn(i, self.propertyDelegate)

    def mouseDoubleClickEvent(self, event) -> None:
        """
        Handles the mouse double click event to open the selected activities.

        Args:
            event: The mouse double click event.
        """
        if self.selected_activities:
            actions.ActivityOpen.run(self.selected_activities)

    @property
    def selected_products(self) -> list[tuple]:
        """
        Returns the selected products.

        Returns:
            list[tuple]: The list of selected products.
        """
        items = [i.internalPointer() for i in self.selectedIndexes() if isinstance(i.internalPointer(), ProductItem)]
        return list({item["key"] for item in items if not item["type"] == "nonfunctional"})

    @property
    def selected_activities(self) -> list[tuple]:
        """
        Returns the selected activities.

        Returns:
            list[tuple]: The list of selected activities.
        """
        items = [i.internalPointer() for i in self.selectedIndexes() if isinstance(i.internalPointer(), ProductItem)]
        return list({item["processor"] if not pd.isna(item["processor"]) else item["key"] for item in items})


class ProductItem(ui.widgets.ABDataItem):
    """
    An item representing a product in the tree view.
    """
    def decorationData(self, col, key):
        """
        Provides decoration data for the item.

        Args:
            col: The column index.
            key: The key for which to provide decoration data.

        Returns:
            The decoration data for the item.
        """
        if key == "name" and self["name"]:
            if self["type"] == "processwithreferenceproduct":
                return ui.icons.qicons.processproduct
            if self["type"] in NODETYPES["biosphere"]:
                return ui.icons.qicons.biosphere
            return ui.icons.qicons.process
        if key == "product":
            if self["type"] in ["product", "processwithreferenceproduct"]:
                return ui.icons.qicons.product
            elif self["type"] == "waste":
                return ui.icons.qicons.waste

    def flags(self, col: int, key: str):
        """
        Returns the item flags for the given column and key.

        Args:
            col (int): The column index.
            key (str): The key for which to return the flags.

        Returns:
            QtCore.Qt.ItemFlags: The item flags.
        """
        return super().flags(col, key) | Qt.ItemFlag.ItemIsDragEnabled

    def displayData(self, col: int, key: str):
        if key.startswith("property_") and not pd.isna(self[key]) and self[key]["normalize"]:
            prop = self[key].copy()
            prop["unit"] = prop['unit'] + f" / {self['unit']}"
            return prop
        return super().displayData(col, key)


class ProductModel(ui.widgets.ABItemModel):
    """
    A model representing the data for the products.

    Attributes:
        dataItemClass (type): The class of the data items.
    """
    dataItemClass = ProductItem

    def mimeData(self, indices: list[QtCore.QModelIndex]):
        """
        Returns the mime data for the given indices.

        Args:
            indices (list[QtCore.QModelIndex]): The indices to get the mime data for.

        Returns:
            core.ABMimeData: The mime data.
        """
        data = core.ABMimeData()
        keys = set(self.values_from_indices("key", indices))
        keys.update(self.values_from_indices("processor", indices))
        keys = {key for key in keys if isinstance(key, tuple)}
        data.setPickleData("application/bw-nodekeylist", list(keys))
        return data

    @staticmethod
    def values_from_indices(key: str, indices: list[QtCore.QModelIndex]):
        """
        Returns the values from the given indices.

        Args:
            key (str): The key to get the values for.
            indices (list[QtCore.QModelIndex]): The indices to get the values for.

        Returns:
            list: The list of values.
        """
        values = []
        for index in indices:
            item = index.internalPointer()
            if not item or item[key] is None:
                continue
            values.append(item[key])
        return values



# -*- coding: utf-8 -*-
import os

# type "localhost:3999" in Chrome for DevTools of AB web content
from activity_browser.utils import get_base_path

os.environ["QTWEBENGINE_REMOTE_DEBUGGING"] = "3999"


def get_static_js_path(file_name: str = "") -> str:
    return str(get_base_path().joinpath("static", "javascript", file_name))


def get_static_css_path(file_name: str = "") -> str:
    return str(get_base_path().joinpath("static", "css", file_name))



from qtpy import QtWidgets, QtGui
from qtpy.QtCore import Qt


class ABDropOverlay(QtWidgets.QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setAttribute(Qt.WA_TransparentForMouseEvents)
        self.setAttribute(Qt.WA_NoSystemBackground)
        self.setAttribute(Qt.WA_TranslucentBackground)
        self.setAutoFillBackground(False)
        self.resize(parent.size())

    def paintEvent(self, event):
        painter = QtGui.QPainter(self)
        painter.setRenderHint(QtGui.QPainter.Antialiasing)
        painter.fillRect(self.rect(), QtGui.QColor(0, 100, 255, 200))  # Semi-transparent blue
        painter.setPen(Qt.white)

        font = self.font()
        font.setBold(True)

        painter.setFont(font)
        painter.drawText(self.rect(), Qt.AlignCenter, "Drop here to create new exchanges")



import re
from collections import namedtuple

import pandas as pd

from asteval import make_symbol_table, Interpreter

from qtpy.QtWidgets import QApplication, QWidget, QCompleter, QTableView, QSizePolicy
from qtpy.QtGui import QPainter, QColor, QFontMetrics, QFontDatabase, QPainterPath, QPen, QFont
from qtpy.QtCore import QTimer, Qt, QAbstractTableModel, QModelIndex

from activity_browser.static import fonts

QFontDatabase.addApplicationFont(fonts.__path__[0] + "/mono.ttf")

operators = r"+\-*/%=<>!&|^~"
pattern = r"\b[a-zA-Z_]\w*\b|[\d.]+|[\"'{}:,+\-*/^()\[\]]| +"


TOKEN_REGEX = r'''
(?P<COMMENT>\#.*)                              # Comments
| (?P<SQSTRING>'(?:\\.|[^\\])*?')              # Single-quoted string
| (?P<DQSTRING>"(?:\\.|[^\\])*?")             # Double-quoted string
| (?P<NUMBER>\b\d+(\.\d*)?([eE][-+]?\d+)?\b)  # Integer or decimal number
| (?P<KEYWORD>\b(?:def|class|if|else|elif|return|for|while|import|from|try|except|with|as|True|False|None|break|continue|pass|and|or|not|is|in|lambda|yield|global|nonlocal|assert|raise|del|async|await)\b)
| (?P<OPERATOR>[+\-*/%=<>!&|^~]+)             # Operators
| (?P<PUNCTUATION>[(),.:;\[\]{}])            # Punctuation
| (?P<IDENTIFIER>\b[a-zA-Z_][a-zA-Z0-9_]*\b)  # Identifiers
| (?P<WHITESPACE>\s+)                         # Whitespace
| (?P<MISMATCH>.)                             # Any other character
'''


def tokenize(expression: str):
    regex = re.compile(TOKEN_REGEX, re.VERBOSE)
    tokens = []
    for match in regex.finditer(expression):
        kind = match.lastgroup
        value = match.group()
        tokens.append(namedtuple("Token", ["kind", "value"])(kind, value))
    return tokens

table = make_symbol_table()

parameters = {
    "db_parameter_1": None,
    "activity_parameter_1": None,
}


class Colors:
    builtin = QColor("#a626a4")
    number = QColor("#986801")
    variable = QColor("#4078f2")
    string = QColor("#50a14f")


class ABFormulaEdit(QWidget):
    def __init__(self, parent=None, scope=None, text=None):
        super().__init__(parent)
        self.scope = scope or {}
        self.error = False
        self.cursor_pos = 0  # Cursor position in the text
        self.selection_start = None  # Selection start index
        self.selection_end = None  # Selection end index
        self.cursor_visible = False  # Blinking cursor state
        self.scroll_offset = 0  # Scroll position for long text
        self.padding = 5  # Left padding for text inside the box
        self.dragging = False  # Track if mouse is dragging

        font = self.font()
        font.setFamily("JetBrains Mono")
        font.setPointSize(9)
        self.setFont(font)

        self.timer = QTimer(self)
        self.timer.timeout.connect(self.toggle_cursor)
        self.timer.start(500)  # Blink cursor every 500ms

        self.interpreter = Interpreter({k: v.amount for k, v in scope.items()})
        self.completer = QCompleter()
        self.completer.setPopup(CompleterView())
        self.completer_model = CompleterModel(self.scope, self.font())

        self.completer.setWidget(self)
        self.completer.setModel(self.completer_model)
        self.completer.setCompletionColumn(0)
        self.completer.activated.connect(self.insert_completion)

        self.text = text or ""  # Stores user input

    @property
    def text(self):
        return self._text

    @text.setter
    def text(self, value):
        self._text = str(value)
        try:
            self.error = None
            self.interpreter.eval(self.text, show_errors=False, raise_errors=True)
        except (SyntaxError, TypeError) as e:
            self.error = e
        except:
            pass

    def pos_to_token_index(self, pos: int) -> int:
        tokens = tokenize(self.text)

        result = -1
        char_index = 0
        for i, (kind, token) in enumerate(tokens):
            char_index += len(token)
            if pos <= char_index:
                result = i
                break

        return result

    def token_at_cursor(self):
        index = self.pos_to_token_index(self.cursor_pos)
        tokens = re.findall(pattern, self.text)

        return tokens[index] if len(tokens) > 0 else ""

    def update_completer(self):
        tokens = tokenize(self.text)
        index = self.pos_to_token_index(self.cursor_pos)

        if len(tokens) == 0 or tokens[index].kind in ["OPERATOR", "PUNCTUATION", "WHITESPACE"]:
            self.completer.setCompletionPrefix("")
        else:
            self.completer.setCompletionPrefix(tokens[index].value)

        self.completer.complete(self.rect())

    def insert_completion(self, completion):
        tokens = tokenize(self.text)
        index = self.pos_to_token_index(self.cursor_pos)

        # if the line is empty, just replace the line with the completion
        if len(tokens) == 0:
            self.text = completion
            self.move_cursor(len(completion))
            return

        text_list = [token.value for token in tokens]

        if tokens[index].kind in ["OPERATOR", "PUNCTUATION", "WHITESPACE"]:
            shift = len(completion)
            text_list.insert(index + 1, completion)
        else:
            shift = len(completion) - len(text_list[index])
            text_list[index] = completion

        self.text = "".join(text_list)
        self.move_cursor(shift)

    def toggle_cursor(self):
        """Toggles cursor visibility for blinking effect."""
        self.cursor_visible = not self.cursor_visible
        self.update()

    def keyPressEvent(self, event):
        """Handles key press events for text input, cursor movement, and selection."""
        key = event.key()
        ctrl_pressed = event.modifiers() & Qt.ControlModifier
        shift_pressed = event.modifiers() & Qt.ShiftModifier

        if ctrl_pressed and key == Qt.Key_C:  # Copy
            self.copy_to_clipboard()
        elif ctrl_pressed and key == Qt.Key_V:  # Paste
            self.paste_from_clipboard()
        elif ctrl_pressed and key == Qt.Key_X:  # Paste
            self.copy_to_clipboard()
            self.delete_selected_text()
        elif ctrl_pressed and key == Qt.Key_A:  # Select All
            self.select_all()
        elif key == Qt.Key_Backspace:
            if self.selection_start is not None:
                self.delete_selected_text()
            elif self.cursor_pos > 0:
                self.text = self.text[:self.cursor_pos - 1] + self.text[self.cursor_pos:]
                self.move_cursor(-1)
        elif key == Qt.Key_Delete:
            if self.selection_start is not None:
                self.delete_selected_text()
            elif self.cursor_pos < len(self.text):
                self.text = self.text[:self.cursor_pos] + self.text[self.cursor_pos + 1:]
        elif key == Qt.Key_Left:
            if shift_pressed:
                self.adjust_selection(-1)
            else:
                self.move_cursor(-1)
        elif key == Qt.Key_Right:
            if shift_pressed:
                self.adjust_selection(1)
            else:
                self.move_cursor(1)
        elif key == Qt.Key_Home:
            self.move_cursor(-self.cursor_pos)
        elif key == Qt.Key_End:
            self.move_cursor(len(self.text) - self.cursor_pos)
        elif key in (Qt.Key_Return, Qt.Key_Enter):
            self.completer.popup().event(event)
        elif event.text():
            if self.selection_start is not None:
                self.delete_selected_text()
            new_text = "".join(re.findall(pattern, event.text()))
            new_text = self.text[:self.cursor_pos] + new_text + self.text[self.cursor_pos:]
            if re.findall(pattern, self.text) == re.findall(pattern, new_text):
                return
            self.text = new_text
            self.move_cursor(1)

        self.adjust_scroll()
        self.update()

    def copy_to_clipboard(self):
        """Copies selected text to the clipboard."""
        if self.selection_start is not None and self.selection_end is not None:
            start, end = sorted([self.selection_start, self.selection_end])
            clipboard = QApplication.clipboard()
            clipboard.setText(self.text[start:end])

    def paste_from_clipboard(self):
        """Pastes text from the clipboard at the cursor position."""
        clipboard = QApplication.clipboard()
        pasted_text = "".join(re.findall(pattern, clipboard.text()))

        if self.selection_start is not None:
            self.delete_selected_text()

        self.text = self.text[:self.cursor_pos] + pasted_text + self.text[self.cursor_pos:]
        self.cursor_pos += len(pasted_text)
        self.adjust_scroll()
        self.update()

    def select_all(self):
        """Selects all text."""
        self.selection_start = 0
        self.selection_end = len(self.text)
        self.cursor_pos = len(self.text)
        self.update()

    def move_cursor(self, step):
        """Moves the cursor without selection."""
        self.cursor_pos = max(0, min(len(self.text), self.cursor_pos + step))
        self.selection_start = None  # Clear selection
        self.adjust_scroll()
        self.update_completer()

    def adjust_selection(self, step):
        """Adjusts selection range while moving the cursor."""
        if self.selection_start is None:
            self.selection_start = self.cursor_pos

        self.cursor_pos = max(0, min(len(self.text), self.cursor_pos + step))
        self.selection_end = self.cursor_pos if self.selection_start != self.cursor_pos else None
        self.adjust_scroll()

    def delete_selected_text(self):
        """Deletes selected text."""
        if self.selection_start is None or self.selection_end is None:
            return
        start, end = sorted([self.selection_start, self.selection_end])
        self.text = self.text[:start] + self.text[end:]
        self.cursor_pos = start
        self.selection_start = self.selection_end = None
        self.adjust_scroll()

    def adjust_scroll(self):
        """Adjusts the scroll position to keep the cursor visible."""
        font_metrics = self.fontMetrics()
        text_width = font_metrics.horizontalAdvance(self.text[:self.cursor_pos])

        if text_width - self.scroll_offset > self.width() - self.padding:
            self.scroll_offset = text_width - self.width() + self.padding
        elif self.scroll_offset and text_width - self.scroll_offset < self.width() - self.padding:
            self.scroll_offset = text_width - self.width() + self.padding
        elif text_width < self.scroll_offset:
            self.scroll_offset = text_width

        if self.scroll_offset < 0:
            self.scroll_offset = 0

    def get_cursor_position_from_x(self, x):
        """Gets the cursor index based on the mouse click position."""
        font_metrics = self.fontMetrics()
        x_offset = x - self.padding + self.scroll_offset
        cursor_pos = len(self.text)

        for i in range(len(self.text)):
            if font_metrics.horizontalAdvance(self.text[:i]) > x_offset:
                cursor_pos = i
                break

        return cursor_pos

    def mousePressEvent(self, event):
        """Handles mouse click events to set cursor position and start selection."""
        if 10 <= event.x() <= 390 and 10 <= event.y() <= 40:
            self.cursor_pos = self.get_cursor_position_from_x(event.x())
            self.selection_start = self.cursor_pos  # Start selection
            self.selection_end = None  # Reset end position
            self.dragging = True  # Start dragging
            self.adjust_scroll()
            self.update()

    def mouseMoveEvent(self, event):
        """Handles mouse dragging for text selection."""
        if self.dragging:
            self.selection_end = self.get_cursor_position_from_x(event.x())
            self.cursor_pos = self.selection_end
            self.adjust_scroll()
            self.update()

    def mouseReleaseEvent(self, event):
        """Stops selection when the mouse is released."""
        self.dragging = False
        if self.selection_end is None or self.selection_end == self.selection_start:
            self.selection_end = self.selection_start = None

    def resizeEvent(self, event):
        super().resizeEvent(event)
        self.adjust_scroll()

    def paintEvent(self, event):
        """Handles drawing the text input field, cursor, and selection."""
        background_color = self.palette().color(self.backgroundRole())

        painter = QPainter(self)
        painter.setPen(Qt.NoPen)
        painter.fillRect(self.rect(), background_color)
        self.paint_text(painter)

    def paint_text(self, painter: QPainter):
        painter.setFont(self.font())

        # Calculate text width and alignment
        font_metrics = painter.fontMetrics()
        text_x = self.padding - self.scroll_offset
        text_y = font_metrics.height() - 2
        cursor_y1 = 0
        cursor_y2 = self.height()
        foreground_color = self.palette().color(self.foregroundRole())

        # Draw selection background if any
        if self.selection_start is not None and self.selection_end is not None:
            start, end = sorted([self.selection_start, self.selection_end])
            start_x = text_x + font_metrics.horizontalAdvance(self.text[:start])
            end_x = text_x + font_metrics.horizontalAdvance(self.text[:end])
            painter.fillRect(start_x, cursor_y1, end_x - start_x, cursor_y2 - cursor_y1,
                             QColor(173, 216, 230))  # Light blue

        # Draw text
        for token_type, token in tokenize(self.text):
            painter.save()

            if not painter.pen() == Qt.NoPen:
                pass

            if token_type == "NUMBER":
                painter.setPen(Colors.number)
            elif token_type in ["SQSTRING", "DQSTRING"]:
                painter.setPen(Colors.string)
            elif token_type == "IDENTIFIER":
                if token in self.scope:
                    painter.setPen(Colors.variable)
                elif token in table:
                    painter.setPen(Colors.builtin)
                else:
                    painter.setPen(foreground_color)
                    draw_error_line(painter, text_x, text_y + 2, font_metrics.horizontalAdvance(token))
            else:
                painter.setPen(foreground_color)

            painter.drawText(text_x, text_y, token)

            text_x += font_metrics.horizontalAdvance(token)
            painter.restore()

        if self.error:
            draw_error_line(painter, self.padding - self.scroll_offset, text_y + 2, text_x)

        # Draw cursor
        if self.cursor_visible and self.selection_start is None:
            cursor_x = self.padding - self.scroll_offset + text_index_to_x(self.text, self.cursor_pos, font_metrics)
            painter.setPen(foreground_color)
            painter.drawLine(cursor_x, cursor_y1, cursor_x, cursor_y2)


class CompleterModel(QAbstractTableModel):
    def __init__(self, data: dict, font: QFont, parent=None):
        super().__init__(parent)
        self.df = pd.DataFrame.from_dict(data, orient="index")

        self._headers = ["Name", "Amount", "Type"]
        self.font = font

    def rowCount(self, parent=QModelIndex()):
        return len(self.df)

    def columnCount(self, parent=QModelIndex()):
        return len(self._headers)

    def data(self, index, role=Qt.DisplayRole):
        if not index.isValid():
            return None

        if role == Qt.DisplayRole:
            if index.column() == 0:
                return self.df.iloc[index.row(), index.column()]
            if index.column() == 1:
                return f"= {self.df.iloc[index.row()]['amount']}"
            if index.column() == 2:
                if self.df.iloc[index.row()]['param_type'] == "project":
                    return "Project Parameter"
                if self.df.iloc[index.row()]['param_type'] == "database":
                    return "Database Parameter"
                if self.df.iloc[index.row()]['param_type'] == "activity":
                    return "Activity Parameter"


        if role == Qt.EditRole:
            return self.df.iloc[index.row(), index.column()]

        if role == Qt.FontRole:
            if index.column() == 0:
                return self.font
            if index.column() == 1:
                font = QFont()
                font.setItalic(True)
                return font
            if index.column() == 2:
                font = QFont()
                font.setPointSize(font.pointSize() - 2)
                return font

        if role == Qt.ForegroundRole and index.column() == 0:
            return Colors.variable

        return None

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        return None


class CompleterView(QTableView):
    def __init__(self):
        super().__init__(showGrid=False)
        self.horizontalHeader().hide()
        self.verticalHeader().hide()
        self.setSizeAdjustPolicy(self.SizeAdjustPolicy.AdjustToContents)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)

    def setModel(self, model):
        super().setModel(model)
        self.resize_on_reset()
        model.modelReset.connect(self.resize_on_reset)

    def resize_on_reset(self):
        self.resizeColumnsToContents()
        for i in range(self.model().rowCount()):
            self.setRowHeight(i, 20)
        self.setFixedWidth(self.sizeHint().width())

def draw_error_line(painter, x, y, w):
    painter.save()
    path = QPainterPath()

    # Define squiggly wave parameters
    amplitude = 2  # Height of the wave
    frequency = 4  # Number of waves across the width

    path.moveTo(x, y)

    for i in range(w // frequency):
        path.lineTo(x + frequency // 2, y - amplitude)
        path.lineTo(x + frequency, y)
        x += frequency

    # Set the error line pen (red squiggly line)
    pen = QPen(Qt.red)
    pen.setWidth(1)
    painter.setPen(pen)

    # Draw the squiggly line
    painter.drawPath(path)
    painter.restore()


def text_index_to_x(text, index, fm: QFontMetrics):
    x = 0
    for token in re.findall(pattern, text):
        if index > len(token):
            x += fm.horizontalAdvance(token)
            index -= len(token)
            continue
        x += fm.horizontalAdvance(token[:index])
        break
    return x


def is_valid_number(string):
    try:
        float(string)  # float will work for both integers and floats
        return True
    except ValueError:
        return False



from logging import getLogger

from qtpy.QtCore import Qt, QEventLoop

from activity_browser import application
from activity_browser.ui import widgets
from activity_browser.actions.base import ABAction, exception_dialogs

log = getLogger(__name__)


class DatabaseOpen(ABAction):
    text = "Open Database"

    @staticmethod
    @exception_dialogs
    def run(database_names: list[str]):
        from activity_browser.layouts import panes

        sibling = DatabaseOpen.find_sibling()

        for db_name in database_names:
            db_pane = panes.DatabaseProductsPane(application.main_window, db_name)
            dock_widget = db_pane.getDockWidget(application.main_window)
            dock_widget.resize(dock_widget.width(), application.main_window.height() // 2)

            application.main_window.addDockWidget(DatabaseOpen.get_area(), dock_widget)

            if sibling:
                application.main_window.tabifyDockWidget(sibling, dock_widget)

                application.thread().eventDispatcher().processEvents(QEventLoop.ProcessEventsFlags.AllEvents)
                dock_widget.raise_()
                dock_widget.show()
            else:
                dock_widget.show()
                application.main_window.resizeDocks(
                    [dock_widget],
                    [1000],
                    Qt.Vertical
                )

    @staticmethod
    def find_sibling():
        """
        Find the dockwidget location where the database pane should be opened.
        """
        from activity_browser.layouts import panes

        all_dws = application.main_window.findChildren(widgets.ABDockWidget)
        databases_dw = application.main_window.findChild(widgets.ABDockWidget, "dockwidget-databases_pane")

        products_dws = [w for w in all_dws if
                        isinstance(w.widget(), panes.DatabaseProductsPane) and
                        application.main_window.dockWidgetArea(w) == application.main_window.dockWidgetArea(databases_dw) and
                        not w.visibleRegion().isNull()
                        ]
        return products_dws[0] if products_dws else None

    @staticmethod
    def get_area():
        """
        Find the dockwidget location where the database pane should be opened.
        """
        databases_dw = application.main_window.findChild(widgets.ABDockWidget, "dockwidget-databases_pane")
        return application.main_window.dockWidgetArea(databases_dw)



from logging import getLogger
from typing import List

from qtpy import QtWidgets

from activity_browser import application
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.ui import widgets
from activity_browser.bwutils import exporters
from activity_browser.ui.core import threading


log = getLogger(__name__)


class DatabaseExportExcel(ABAction):
    """
    ABAction to export database(s) to Excel format (.xlsx).
    """

    icon = application.style().standardIcon(QtWidgets.QStyle.SP_DriveHDIcon)
    text = "Export to Excel (.xlsx)"
    tool_tip = "Export database(s) to Excel format"

    @classmethod
    @exception_dialogs
    def run(cls, db_names: List[str] = None):
        if db_names is None:
            import bw2data as bd
            dialog = widgets.ABDatabaseSelectionDialog(
                parent=application.main_window,
                databases=sorted(bd.databases),
                title="Select databases to export to Excel"
            )
            if dialog.exec_() == QtWidgets.QDialog.Accepted:
                db_names = dialog.get_selected_databases()
            else:
                return

        # Get export directory or file from user
        if len(db_names) == 1:
            # Single database - suggest a filename
            suggested_name = f"lci-{db_names[0]}.xlsx"
            path, _ = QtWidgets.QFileDialog.getSaveFileName(
                parent=application.main_window,
                caption=f'Export database "{db_names[0]}" to Excel',
                directory=suggested_name,
                filter='Excel spreadsheet (*.xlsx);; All files (*.*)'
            )
        else:
            # Multiple databases - ask for directory
            path = QtWidgets.QFileDialog.getExistingDirectory(
                parent=application.main_window,
                caption=f'Select directory to export {len(db_names)} databases',
            )
        
        if not path:
            return

        # Show export dialog
        context = {
            "db_names": db_names,
            "path": path,
        }
        export_dialog = ExportExcelSetup(
            parent=application.main_window,
            title="Export to Excel",
            context=context
        )
        export_dialog.show()


class ExportExcelSetup(widgets.ABWizard):
    """Wizard for exporting databases to Excel format."""

    class ExportPage(widgets.ABThreadedWizardPage):
        """Wizard page to export the selected database(s) to Excel."""
        title = "Exporting Database(s)"
        subtitle = "Exporting database(s) to Excel file(s)"

        class Thread(threading.ABThread):
            """Thread to handle the export process."""
            
            def run_safely(self, db_names: List[str], path: str):
                """Export the database(s) to Excel."""
                for db_name in db_names:
                    try:
                        exporters.write_lci_excel(db_name, path)
                        log.info(f"Successfully exported database '{db_name}' to Excel")
                    except Exception as e:
                        log.error(f"Failed to export database '{db_name}': {e}")
                        raise

        def initializePage(self, context: dict):
            """Start the export thread."""
            self.thread.start(context["db_names"], context["path"])

    pages = [ExportPage]



from logging import getLogger

from qtpy import QtGui

from activity_browser import signals
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.mod import bw2data as bd

log = getLogger(__name__)


class DatabaseRedoAllocation(ABAction):
    """
    ABAction to redo the allocation calculation.
    """

    icon = QtGui.QIcon()
    text = "Redo allocation for database"
    tool_tip = "Redo the allocation calculations for this database"

    @staticmethod
    @exception_dialogs
    def run(db_name: str):
        if bd.databases[db_name].get("backend") == "multifunctional":
            try:
                db = bd.Database(db_name)

                for node in filter(lambda x: x.multifunctional, db):
                    node.allocate()

                signals.new_statusbar_message.emit(f"Allocation values for database {db_name} updated.")
            except KeyError as exc:
                signals.new_statusbar_message.emit("A property for the allocation calculation was not found!")
                log.error(f"A property for the allocation calculation was not found: {exc}")
            except ZeroDivisionError as exc:
                signals.new_statusbar_message.emit(str(exc))
                log.error(f"Zero division in allocation calculation: {exc}")



from importlib.metadata import version

import bw2data as bd

from qtpy import QtGui, QtWidgets
from qtpy.QtCore import QSize, QUrl

from activity_browser import actions, signals, utils, application

from .icons import qicons


class MenuBar(QtWidgets.QMenuBar):
    """
    Main menu bar at the top of the Activity Browser window. Contains submenus for different user interaction categories
    """
    def __init__(self, window):
        super().__init__(parent=window)

        self.project_menu = ProjectMenu(self)
        self.view_menu = ViewMenu(self)
        self.calculate_menu = CalculateMenu(self)
        # self.tools_menu = ToolsMenu(self)
        self.help_menu = HelpMenu(self)

        self.addMenu(self.project_menu)
        self.addMenu(self.view_menu)
        self.addMenu(self.calculate_menu)
        # self.addMenu(self.tools_menu)
        self.addMenu(self.help_menu)


class ProjectMenu(QtWidgets.QMenu):
    """
    Project menu: contains actions related to managing the project, such as project duplication, database importing etc.
    """

    def __init__(self, parent=None) -> None:
        super().__init__(parent)

        self.setTitle("&Project")

        self.dup_proj_action = actions.ProjectDuplicate.get_QAction()
        self.delete_proj_action = actions.ProjectDelete.get_QAction()

        self.import_proj_action = actions.ProjectImport.get_QAction()
        self.export_proj_action = actions.ProjectExport.get_QAction()

        self.manage_settings_action = actions.SettingsWizardOpen.get_QAction()
        self.manage_projects_action = actions.ProjectManagerOpen.get_QAction()

        self.addMenu(ProjectSelectionMenu(self))
        self.addMenu(ProjectNewMenu(self))
        self.addAction(self.dup_proj_action)
        self.addAction(self.delete_proj_action)
        self.addSeparator()
        self.addAction(self.import_proj_action)
        self.addAction(self.export_proj_action)
        self.addSeparator()
        self.addMenu(ImportDatabaseMenu(self))
        self.addMenu(ExportDatabaseMenu(self))
        self.addSeparator()
        self.addMenu(ImportICMenu(self))
        self.addSeparator()
        self.addAction(self.manage_settings_action)
        self.addAction(self.manage_projects_action)


class ProjectNewMenu(QtWidgets.QMenu):
    def __init__(self, parent=None) -> None:
        super().__init__(parent)

        self.setTitle("New project")
        self.new_proj_action = actions.ProjectNew.get_QAction()
        self.import_proj_action = actions.ProjectImport.get_QAction()

        self.new_proj_action.setText("Empty project")
        self.import_proj_action.setText("From .tar.gz file")

        self.new_proj_action.setIcon(QtGui.QIcon())
        self.import_proj_action.setIcon(QtGui.QIcon())

        self.addAction(self.new_proj_action)
        self.addAction(self.import_proj_action)
        self.addMenu(ProjectNewTemplateMenu(self))


class ProjectNewTemplateMenu(QtWidgets.QMenu):
    remote_projects = {}

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setTitle("From template")

        self.actions = {}

        for key in utils.get_templates():
            action = actions.ProjectNewFromTemplate.get_QAction(key)
            action.setText(key)
            self.actions[key] = action
            self.addAction(action)

        for key in self.get_projects():
            action = actions.ProjectNewRemote.get_QAction(key)
            action.setText(key)
            self.actions[key] = action
            self.addAction(action)

    def get_projects(self):
        if not self.remote_projects:
            from bw2io.remote import get_projects
            ProjectNewTemplateMenu.remote_projects = get_projects()
        return self.remote_projects


class ViewMenu(QtWidgets.QMenu):
    """
    View menu: contains actions in regard to hiding and showing specific UI elements.
    """

    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        self.setTitle("&View")


class CalculateMenu(QtWidgets.QMenu):
    """
    Calculate Menu: contains actions in regard to calculating the LCA results for the current project
    """

    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        self.setTitle("&Calculate")
        self.cs_actions = []

        self.new_cs_action = actions.CSNew.get_QAction()
        self.new_cs_action.setText("New setup...")
        self.addAction(self.new_cs_action)
        self.addSeparator()

        signals.project.changed.connect(self.sync)
        signals.meta.calculation_setups_changed.connect(self.sync)

    def sync(self):
        self.cs_actions.clear()
        for cs in bd.calculation_setups:
            action = actions.CSOpen.get_QAction(cs)
            action.setText(cs)
            self.cs_actions.append(action)
            self.addAction(action)


# class ToolsMenu(QtWidgets.QMenu):
#     """
#     Tools Menu: contains actions in regard to special tooling aspects of the AB
#     """
#
#     def __init__(self, parent=None) -> None:
#         super().__init__(parent)
#         self.setTitle("&Tools")
#
#         self.manage_plugins_action = actions.PluginWizardOpen.get_QAction()
#
#         self.addAction(self.manage_plugins_action)


class HelpMenu(QtWidgets.QMenu):
    """
    Help Menu: contains actions that show info to the user or redirect them to online resources
    """

    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        self.setTitle("&Help")

        self.addAction(
            qicons.ab, "&About Activity Browser", self.about
        )
        self.addAction(
            "&About Qt", lambda: QtWidgets.QMessageBox.aboutQt(application.main_window)
        )
        self.addAction(
            qicons.question, "&Get help on the wiki", self.open_wiki
        )
        self.addAction(
            qicons.issue, "&Report an idea/issue on GitHub", self.raise_issue_github
        )

    def about(self):
        """Displays an 'about' window to the user containing e.g. the version of the AB and copyright info"""
        # set the window text in html format
        text = f"""
        Activity Browser - a graphical interface for Brightway2.<br><br>
        Application version: <b>{version("activity_browser")}</b><br>
        bw2data version: <b>{version("bw2data")}</b><br>
        bw2io version: <b>{version("bw2calc")}</b><br>
        bw2calc version: <b>{version("bw2io")}</b><br><br>
        All development happens on <a href="https://github.com/LCA-ActivityBrowser/activity-browser">github</a>.<br><br>
        For copyright information please see the copyright on <a href="https://github.com/LCA-ActivityBrowser/activity-browser/tree/main#copyright">this page</a>.<br><br>
        For license information please see the copyright on <a href="https://github.com/LCA-ActivityBrowser/activity-browser/blob/main/LICENSE.txt">this page</a>.<br><br>
        """

        # set up the window
        about_window = QtWidgets.QMessageBox(parent=application.main_window)
        about_window.setWindowTitle("About the Activity Browser")
        about_window.setIconPixmap(qicons.ab.pixmap(QSize(150, 150)))
        about_window.setText(text)

        # execute
        about_window.exec_()

    def open_wiki(self):
        """Opens the AB github wiki in the users default browser"""
        url = QUrl(
            "https://github.com/LCA-ActivityBrowser/activity-browser/wiki"
        )
        QtGui.QDesktopServices.openUrl(url)

    def raise_issue_github(self):
        """Opens the github create issue page in the users default browser"""
        url = QUrl(
            "https://github.com/LCA-ActivityBrowser/activity-browser/issues/new/choose"
        )
        QtGui.QDesktopServices.openUrl(url)


class ProjectSelectionMenu(QtWidgets.QMenu):
    """
    Menu that lists all the projects available through bw2data.projects
    """
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setTitle("Open project")
        self.populate()

        self.aboutToShow.connect(self.populate)
        self.triggered.connect(lambda act: actions.ProjectSwitch.run(act.data()))

    def populate(self):
        """
        Populates the menu with the projects available in the database
        """
        import bw2data as bd

        # clear the menu of any already existing actions
        self.clear()

        # sort projects alphabetically
        sorted_projects = sorted(list(bd.projects))

        # iterate over the sorted projects and add them as actions to the menu
        for i, proj in enumerate(sorted_projects):
            # check whether the project is BW25
            bw_25 = (
                False if not isinstance(proj.data, dict) else proj.data.get("25", False)
            )

            # create the action and disable it if it's BW25 and BW25 is not supported
            action = QtWidgets.QAction(proj.name, self)
            action.setData(proj.name)
            action.setIcon(
                application.style().standardIcon(QtWidgets.QStyle.StandardPixmap.SP_MessageBoxWarning) if not bw_25 else qicons.empty)

            self.addAction(action)


class ImportDatabaseMenu(QtWidgets.QMenu):
    def __init__(self, parent=None) -> None:
        super().__init__(parent=parent)
        self.setTitle("Import database")
        self.setIcon(qicons.import_db)

        self.import_from_ecoinvent_action = actions.DatabaseImportFromEcoinvent.get_QAction()
        self.import_from_excel_action = actions.DatabaseImporterExcel.get_QAction()
        self.import_from_bw2package_action = actions.DatabaseImporterBW2Package.get_QAction()

        self.import_from_ecoinvent_action.setText("ecoinvent...")
        self.import_from_excel_action.setText("from .xlsx")
        self.import_from_bw2package_action.setText("from .bw2package")

        self.addAction(self.import_from_excel_action)
        self.addAction(self.import_from_bw2package_action)
        self.addSeparator()
        self.addAction(self.import_from_ecoinvent_action)


class ExportDatabaseMenu(QtWidgets.QMenu):
    def __init__(self, parent=None) -> None:
        super().__init__(parent=parent)
        self.setTitle("Export database")

        self.export_to_excel_action = actions.DatabaseExportExcel.get_QAction()
        self.export_to_bw2package_action = actions.DatabaseExportBW2Package.get_QAction()

        self.export_to_excel_action.setText("to .xlsx")
        self.export_to_bw2package_action.setText("to .bw2package")

        self.addAction(self.export_to_excel_action)
        self.addAction(self.export_to_bw2package_action)


class ImportICMenu(QtWidgets.QMenu):
    def __init__(self, parent=None) -> None:
        super().__init__(parent=parent)
        self.setTitle("Import impact categories")
        self.setIcon(qicons.import_db)

        self.beta_warning = QtWidgets.QWidgetAction(self)
        self.beta_warning.setDefaultWidget(QtWidgets.QLabel("Beta features, use at your own risk"))

        self.import_from_ei_excel_action = actions.MethodImporterEcoinvent.get_QAction()
        self.import_from_bw2io_action = actions.MethodImporterBW2IO.get_QAction()

        self.import_from_ei_excel_action.setText("from ecoinvent excel")
        self.import_from_bw2io_action.setText("from bw2io")

        self.import_from_ei_excel_action.setIcon(QtGui.QIcon())
        self.import_from_bw2io_action.setIcon(QtGui.QIcon())

        self.addAction(self.beta_warning)
        self.addSeparator()
        self.addAction(self.import_from_ei_excel_action)
        self.addAction(self.import_from_bw2io_action)



# -*- coding: utf-8 -*-
from qtpy import QtCore, QtWidgets


class StringDelegate(QtWidgets.QStyledItemDelegate):
    """For managing and validating entered string values."""

    def displayText(self, value, locale):
        if isinstance(value, (list, tuple)):
            return ", ".join(value)

        return str(value)

    def createEditor(self, parent, option, index):
        editor = QtWidgets.QLineEdit(parent)
        return editor

    def setEditorData(self, editor: QtWidgets.QLineEdit, index: QtCore.QModelIndex):
        """Populate the editor with data if editing an existing field."""
        value = index.data(QtCore.Qt.DisplayRole)
        # Avoid setting 'None' type value as a string
        value = str(value) if value else ""
        editor.setText(value)

    def setModelData(
        self,
        editor: QtWidgets.QLineEdit,
        model: QtCore.QAbstractItemModel,
        index: QtCore.QModelIndex,
    ):
        """Take the editor, read the given value and set it in the model"""
        value = editor.text()
        model.setData(index, value, QtCore.Qt.EditRole)



import py7zr
import multiprocessing
import tempfile
import os
from io import BytesIO
from lxml import objectify
from functools import partial
from logging import getLogger

import tqdm
import bw2data as bd
from bw2io.extractors.ecospold2 import getattr2, ACTIVITY_TYPES, Ecospold2DataExtractor
from bw2io.importers.ecospold2_biosphere import Ecospold2BiosphereImporter
from bw2io.strategies import (
    add_cpc_classification_from_single_reference_product,
    assign_single_product_as_activity,
    convert_activity_parameters_to_list,
    create_composite_code,
    delete_exchanges_missing_activity,
    delete_ghost_exchanges,
    delete_none_synonyms,
    drop_temporary_outdated_biosphere_flows,
    drop_unspecified_subcategories,
    es2_assign_only_product_with_amount_as_reference_product,
    fix_ecoinvent_flows_pre35,
    fix_unreasonably_high_lognormal_uncertainties,
    link_biosphere_by_flow_uuid,
    link_internal_technosphere_by_composite_code,
    normalize_units,
    remove_uncertainty_from_negative_loss_exchanges,
    remove_unnamed_parameters,
    remove_zero_amount_coproducts,
    remove_zero_amount_inputs_with_no_activity,
    update_ecoinvent_locations,
    update_social_flows_in_older_consequential,
)

log = getLogger(__name__)


class Ecoinvent7zImporter:

    def __init__(self, archive_path: str) -> None:
        self.archive_path = archive_path
        self.is_compressed = archive_path.endswith('.7z')

    def install_biosphere(self, biosphere_name: str = "biosphere3") -> None:
        """
        Installs the biosphere that is bundled in the ecoinvent .7z. Simple extraction and installing through the
        standard bw2io importer is quick enough because we're talking about a single file.
        """
        if self.is_compressed:
            # extract the elementary exchanges to a temporary location
            with py7zr.SevenZipFile(self.archive_path, mode='r') as archive:
                temp = tempfile.gettempdir()
                archive.extract(temp, ["MasterData/ElementaryExchanges.xml"])
                bio_file = os.path.join(temp, "MasterData", "ElementaryExchanges.xml")
        else:
            bio_file = os.path.join(self.archive_path, "MasterData", "ElementaryExchanges.xml")

        # initiate the importer with the elementary exchanges file
        importer = Ecospold2BiosphereImporter(biosphere_name, None, bio_file)

        # install the biosphere
        importer.apply_strategies()
        importer.write_database(searchable=False)

    def install_ecoinvent(self, db_name, biosphere_name: str = "biosphere3"):
        """
        Runs the importer by reading the .7z archive into memory and multiprocessing it from there. This is faster than
        the usual method which is very much IO-bound.
        """
        # if the db already exists, warn the user of the impending overwriting and delete the existing database
        if db_name in bd.databases:
            log.warning(f"Database already exists, overwriting {db_name}")
            bd.Database(db_name).delete(warn=False)

        if self.is_compressed:
            # load the spold files into memory
            spold_bytes = self.read_archive_to_bytes()

            # process the in-memory data to a dict format and applying strategies
            db_data = self.process_bytes(spold_bytes, db_name)
        else:
            # if the archive is not compressed, we can use the standard importer
            db_data = Ecospold2DataExtractor.extract(os.path.join(self.archive_path, "datasets"), db_name)

        db_data = self.apply_strategies(db_data, biosphere_name)

        # rewrite into an ingestible format
        db_data = {(ds["database"], ds["code"]): ds for ds in db_data}

        # creating the actual database
        db = bd.Database(db_name)
        db.write(db_data, searchable=False)

    def apply_strategies(self, db_data, biosphere_name):
        strategies = [
            normalize_units,
            update_ecoinvent_locations,
            remove_zero_amount_coproducts,
            remove_zero_amount_inputs_with_no_activity,
            remove_unnamed_parameters,
            es2_assign_only_product_with_amount_as_reference_product,
            assign_single_product_as_activity,
            create_composite_code,
            drop_unspecified_subcategories,
            fix_ecoinvent_flows_pre35,
            drop_temporary_outdated_biosphere_flows,
            partial(link_biosphere_by_flow_uuid, biosphere=biosphere_name),
            link_internal_technosphere_by_composite_code,
            delete_exchanges_missing_activity,
            delete_ghost_exchanges,
            remove_uncertainty_from_negative_loss_exchanges,
            fix_unreasonably_high_lognormal_uncertainties,
            convert_activity_parameters_to_list,
            add_cpc_classification_from_single_reference_product,
            delete_none_synonyms,
            partial(update_social_flows_in_older_consequential, biosphere_db=bd.Database(biosphere_name)),
        ]
        for strategy in tqdm.tqdm(strategies, desc="Applying strategies", total=len(strategies)):
            db_data = strategy(db_data)
        return db_data

    def read_archive_to_bytes(self) -> {str: BytesIO}:
        log.info("Extracting .7z archive to memory")
        with py7zr.SevenZipFile(self.archive_path, mode='r') as archive:
            # Find all .spold dataset files
            file_list = [
                file.filename for file in archive.list()
                if file.filename.startswith("datasets")
                and file.filename.endswith(".spold")
            ]

            extracted_data = archive.read(file_list)

        return extracted_data

    def process_bytes(self, spold_bytes: {str: BytesIO}, db_name: str) -> list:
        with multiprocessing.Pool(processes=multiprocessing.cpu_count()) as pool:
            log.info(f"Extracting XML data from {len(spold_bytes)} datasets")
            results = [
                pool.apply_async(
                    self.extract_activity,
                    args=(x, db_name),
                )
                for x in spold_bytes.values()
            ]

            data = []
            for result in tqdm.tqdm(results, desc="Processing datasets", total=len(results)):
                data.append(result.get())

        return data

    @staticmethod
    def extract_activity(xml_bytes, db_name):
        root = objectify.parse(xml_bytes).getroot()
        if hasattr(root, "activityDataset"):
            stem = root.activityDataset
        else:
            stem = root.childActivityDataset

        comments = [
            Ecospold2DataExtractor.condense_multiline_comment(
                getattr2(stem.activityDescription.activity, "generalComment")
            ),
            (
                "Included activities start: ",
                getattr2(
                    stem.activityDescription.activity, "includedActivitiesStart"
                ).get("text"),
            ),
            (
                "Included activities end: ",
                getattr2(
                    stem.activityDescription.activity, "includedActivitiesEnd"
                ).get("text"),
            ),
            (
                "Geography: ",
                Ecospold2DataExtractor.condense_multiline_comment(
                    getattr2(stem.activityDescription.geography, "comment")
                ),
            ),
            (
                "Technology: ",
                Ecospold2DataExtractor.condense_multiline_comment(
                    getattr2(stem.activityDescription.technology, "comment")
                ),
            ),
            (
                "Time period: ",
                Ecospold2DataExtractor.condense_multiline_comment(
                    getattr2(stem.activityDescription.timePeriod, "comment")
                ),
            ),
        ]
        comment = "\n".join(
            [
                (" ".join(x) if isinstance(x, tuple) else x)
                for x in comments
                if (x[1] if isinstance(x, tuple) else x)
            ]
        )

        classifications = [
            (el.classificationSystem.text, el.classificationValue.text)
            for el in stem.activityDescription.iterchildren()
            if el.tag == u"{http://www.EcoInvent.org/EcoSpold02}classification"
        ]

        data = {
            "comment": comment,
            "classifications": classifications,
            "activity type": ACTIVITY_TYPES[
                int(stem.activityDescription.activity.get("specialActivityType") or 0)
            ],
            'activity': stem.activityDescription.activity.get('id'),
            'database': db_name,
            "exchanges": [
                Ecospold2DataExtractor.extract_exchange(exc)
                for exc in stem.flowData.iterchildren()
                if "parameter" not in exc.tag
            ],
            # 'filename':  os.path.basename(filename),
            'location': stem.activityDescription.geography.shortname.text,
            'name': stem.activityDescription.activity.activityName.text,
            'synonyms': [s.text for s in getattr(stem.activityDescription.activity, 'synonym', [])],
            "parameters": dict(
                [
                    Ecospold2DataExtractor.extract_parameter(exc)
                    for exc in stem.flowData.iterchildren()
                    if "parameter" in exc.tag
                ]
            ),
            "authors": {
                "data entry": {
                    "name": stem.administrativeInformation.dataEntryBy.get(
                        "personName"
                    ),
                    "email": stem.administrativeInformation.dataEntryBy.get(
                        "personEmail"
                    ),
                },
                "data generator": {
                    "name": stem.administrativeInformation.dataGeneratorAndPublication.get(
                        "personName"
                    ),
                    "email": stem.administrativeInformation.dataGeneratorAndPublication.get(
                        "personEmail"
                    ),
                },
            },
            "type": "process",
        }
        return data




from logging import getLogger

import pandas as pd
from qtpy import QtWidgets, QtCore

import bw2data as bd
from bw2io import remote

from activity_browser import actions, ui, signals, utils
from activity_browser.settings import ab_settings
from activity_browser.ui import widgets


log = getLogger(__name__)


class ProjectManagerPane(widgets.ABAbstractPane):
    title = "Project Manager"
    unique = True

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Project Manager")

        self.tabs = QtWidgets.QTabWidget(self)

        self.project_model = widgets.ABItemModel(self)
        self.project_model.dataItemClass = ProjectItem

        self.template_model = widgets.ABItemModel(self)
        self.template_model.dataItemClass = TemplateItem

        self.project_view = ProjectView(self)
        self.project_view.setModel(self.project_model)

        self.template_view = TemplateView(self)
        self.template_view.setModel(self.template_model)

        self.sync()

        self.tabs.addTab(self.project_view, "Projects")
        self.tabs.addTab(self.template_view, "Templates")

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.tabs)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

        # connect signals
        signals.project.changed.connect(self.sync)
        signals.project.deleted.connect(self.sync)

    def sync(self):
        self.project_model.setDataFrame(self.build_project_df())
        self.template_model.setDataFrame(self.build_template_df())

    def build_project_df(self) -> pd.DataFrame:
        data = {}
        for proj_ds in sorted(bd.projects):
            # if for any reason the project data is not a dictionary, log a warning and set it to an empty dict
            if not isinstance(proj_ds.data, dict):
                log.warning(f"Project {proj_ds.name} has no data dictionary")
                proj_ds.data = {}

            data[proj_ds.name] = {
                "Name": proj_ds.name,
                "Path": proj_ds.dir,
                "Version": "Brightway25" if proj_ds.data.get("25", False) else "Legacy"
            }

        return pd.DataFrame.from_dict(data, orient="index")

    def build_template_df(self) -> pd.DataFrame:
        data = {}

        templates = utils.get_templates()
        remote_templates = remote.get_projects()

        for name in sorted(templates):
            data[name] = {
                "Name": name,
                "Path": templates[name],
                "Remote": "No"
            }

        for name in sorted(remote_templates):
            data[name] = {
                "Name": name,
                "Path": remote_templates[name],
                "Remote": "Yes"
            }

        return pd.DataFrame.from_dict(data, orient="index")


class ProjectView(widgets.ABTreeView):

    class ContextMenu(widgets.ABTreeView.ContextMenu):
        def __init__(self, pos, view: "FunctionView"):
            from activity_browser.ui.menu_bar import ProjectNewMenu

            super().__init__(pos, view)
            items = list({index.internalPointer() for index in view.selectedIndexes()})

            self.addMenu(ProjectNewMenu(self))

            if len(items) == 0:
                return

            if len(items) == 1:
                self.dup_project = actions.ProjectDuplicate.get_QAction(items[0]["Name"])
                self.template_project = actions.ProjectCreateTemplate.get_QAction(items[0]["Name"], view.parent())
                self.addAction(self.dup_project)
                self.addAction(self.template_project)

            if len(items) == 1 and len([i for i in items if i["Version"] == "Legacy"]) == 1:
                self.migrate_project = actions.ProjectMigrate25.get_QAction(items[0]["Name"])
                self.addAction(self.migrate_project)

            self.del_project = actions.ProjectDelete.get_QAction(view.selected_projects)
            self.addAction(self.del_project)


    def __init__(self, parent: ProjectManagerPane):
        super().__init__(parent)
        self.setSortingEnabled(True)
        self.setSelectionBehavior(ui.widgets.ABTreeView.SelectionBehavior.SelectRows)
        self.setSelectionMode(ui.widgets.ABTreeView.SelectionMode.ExtendedSelection)


    @property
    def selected_projects(self) -> [str]:
        items = [i.internalPointer() for i in self.selectedIndexes() if isinstance(i.internalPointer(), ProjectItem)]
        return list({item["Name"] for item in items if item["Name"] is not None})


class ProjectItem(widgets.ABDataItem):
    def decorationData(self, col, key):
        if col != 0:
            return
        return ui.icons.qicons.forward if self["Name"] == ab_settings.startup_project else ui.icons.QIcons.forward


class TemplateView(widgets.ABTreeView):

    class ContextMenu(widgets.ABTreeView.ContextMenu):
        def __init__(self, pos, view: "FunctionView"):
            super().__init__(pos, view)

            items = list({index.internalPointer() for index in view.selectedIndexes()})

    def __init__(self, parent: ProjectManagerPane):
        super().__init__(parent)
        self.setSortingEnabled(True)
        self.setSelectionBehavior(ui.widgets.ABTreeView.SelectionBehavior.SelectRows)


class TemplateItem(widgets.ABDataItem):
    def decorationData(self, col, key):
        if col != 0:
            return
        return ui.icons.qicons.forward if self["Name"] == ab_settings.startup_project else ui.icons.QIcons.forward



from qtpy import QtWidgets, QtCore, QtGui
from qtpy.QtCore import Qt


class HideMode:
    Close = 1
    Hide = 2


class ABDockWidget(QtWidgets.QDockWidget):
    updatingTabBar = False
    HideMode = HideMode

    def __init__(self, title: str, parent: QtWidgets.QMainWindow, mode=HideMode.Close) -> None:
        super().__init__(title, parent)
        self.main = parent
        self.title = title

        self._updating_tab_bar = ""

        self._hide_mode = mode
        if self._hide_mode == HideMode.Close:
            self.setAttribute(Qt.WA_DeleteOnClose)

        self.title_bar = TitleBar(title, self.button(), self)
        self.setTitleBarWidget(self.title_bar)

    def setWidget(self, widget):
        super().setWidget(widget)
        widget.destroyed.connect(self.deleteLater)
        self.setObjectName(f"dockwidget-{widget.objectName()}")

    def button(self):
        if self._hide_mode == HideMode.Close:
            button = CloseButton(self)
            button.clicked.connect(self.close)
        else:
            button = MinimizeButton(self)
            button.clicked.connect(self.hide)
        return button


class TitleBar(QtWidgets.QWidget):
    def __init__(self, title: str, button, parent=None):
        super().__init__(parent)

        self.label = QtWidgets.QLabel(title, self)
        font = self.font()
        font.setPointSize(font.pointSize() + 1)
        self.label.setFont(font)
        self.label.setCursor(QtGui.QCursor(Qt.CursorShape.SizeAllCursor))

        layout = QtWidgets.QHBoxLayout()
        layout.addWidget(self.label)
        layout.addStretch()
        layout.addWidget(button)
        self.setLayout(layout)

    def set_button(self, button):
        layout = self.layout()
        w = layout.itemAt(2).widget()
        layout.replaceWidget(w, button)
        layout.update()
        w.deleteLater()


class CloseButton(QtWidgets.QWidget):
    """Custom close button with hover effect."""
    clicked: QtCore.SignalInstance = QtCore.Signal()

    def __init__(self, parent=None):
        super().__init__(parent)


        self.label = QtWidgets.QLabel("×", self)

        self.label.setFont(QtGui.QFont("Arial", 12, QtGui.QFont.Bold))
        self.label.setAlignment(Qt.AlignCenter)
        self.label.setFixedSize(16, 16)
        self.label.mousePressEvent = lambda event: self.clicked.emit()

        self.label.setStyleSheet("""
            QLabel {
                border-radius: 8px;
                background-color: transparent;
            }
            QLabel:hover {
                background-color: rgba(255, 0, 0, 0.5);
            }
        """)

        layout = QtWidgets.QHBoxLayout()
        layout.setContentsMargins(5, 0, 0, 0)
        layout.addWidget(self.label)
        self.setLayout(layout)


class MinimizeButton(QtWidgets.QWidget):
    """Custom close button with hover effect."""
    clicked: QtCore.SignalInstance = QtCore.Signal()

    def __init__(self, parent=None):
        super().__init__(parent)

        self.label = QtWidgets.QLabel("-", self)

        self.label.setFont(QtGui.QFont("Arial", 12, QtGui.QFont.Bold))
        self.label.setAlignment(Qt.AlignCenter)
        self.label.setFixedSize(16, 16)
        self.label.mousePressEvent = lambda event: self.clicked.emit()

        self.setStyleSheet("""
            QLabel {
                border-radius: 8px;
                background-color: transparent;
            }
            QLabel:hover {
                background-color: rgba(42, 157, 244, 0.5);
            }
        """)

        layout = QtWidgets.QHBoxLayout()
        layout.setContentsMargins(5, 0, 0, 0)
        layout.addWidget(self.label)
        self.setLayout(layout)


def mousePressEvent(self, event):
    if event.button() == Qt.LeftButton:
        self.drag_start_pos = event.pos()


def mouseMoveEvent(self, event):
    if not self.drag_start_pos:
        return

    # Check if mouse moved beyond threshold
    if (event.pos() - self.drag_start_pos).manhattanLength() > QtWidgets.QApplication.startDragDistance():
        index = self.tabAt(self.drag_start_pos)
        if index >= 0:
            startDrag(self, index)

def startDrag(self, index):
    """Start dragging a tab."""
    print("Dragging success")



from logging import getLogger

import pandas as pd
import numpy as np
import timeit

from qtpy import QtCore

from activity_browser import signals, application

from .metadata import MetaDataStore
from .fields import primary, secondary, all_types

log = getLogger(__name__)


class MDSUpdater(QtCore.QObject):
    def __init__(self, mds: MetaDataStore):
        super().__init__(mds)

        self.mds = mds
        self.connect_signals()

    def connect_signals(self):
        signals.node.changed.connect(self.on_node_changed)
        signals.node.deleted.connect(self.on_node_deleted)

        signals.meta.databases_changed.connect(self.on_database_changed)
        signals.database.deleted.connect(self.on_database_changed)

    # callbacks
    def on_node_changed(self, new, old):
        node_data = {f: getattr(new, f) for f in primary}
        node_data = node_data | {f: new.data.get(f, np.NaN) for f in secondary}
        node_data["key"] = new.key
        node_data = pd.Series(node_data, name=new.key)

        if new.key in self.mds.dataframe.index and not all(node_data.dropna().eq(self.mds.dataframe.loc[new.key].dropna())):
            self.modify_node(node_data)
        else:
            self.add_node(node_data)

    def on_node_deleted(self, ds):
        try:
            self.delete_node(ds)
        except KeyError:
            pass

    def on_database_changed(self) -> None:
        databases = databases_in_sqlite()

        for db_name in [x for x in self.mds.databases if x not in databases]:
            self.delete_database(db_name)

        for db_name in [x for x in databases if x not in self.mds.databases]:
            self.add_database(db_name)

    # node methods
    def modify_node(self, ds: pd.Series):
        self._fix_categories(ds)
        self.mds.dataframe.loc[ds.key] = ds
        self.mds.register_mutation(ds.key, "update")

    def add_node(self, ds: pd.Series):
        self._fix_categories(ds)
        self.mds.dataframe.loc[ds.key, :] = ds
        self.mds.register_mutation(ds.key, "add")

    def delete_node(self, ds: pd.Series):
        self.mds.dataframe = self.mds.dataframe.drop(ds.key)
        self.mds.register_mutation(ds.key, "delete")

    # database methods
    def add_database(self, db_name: str):
        self.mds.loader.load_database(db_name)

    def delete_database(self, db_name: str):
        for code in self.mds.dataframe.loc[db_name].index:
            self.mds.register_mutation((db_name, code), "delete")

        self.mds.dataframe = self.mds.dataframe.drop(db_name, level=0)

    # utility functions
    def _fix_categories(self, ds: pd.Series):
        for category_col in [k for k, v in all_types.items() if k in ds and v == "category"]:
            category = ds[category_col]

            if pd.isna(category):
                # cannot add NaN as a category
                continue

            if category in self.mds.dataframe[category_col].cat.categories:
                # category already exists
                continue

            # add new category to column
            self.mds.dataframe[category_col] = self.mds.dataframe[category_col].cat.add_categories([category])



def databases_in_sqlite() -> set[str]:
    import sqlite3
    from bw2data.backends import sqlite3_lci_db

    with sqlite3.connect(sqlite3_lci_db._filepath) as db:
        cursor = db.cursor()
        result = cursor.execute("SELECT DISTINCT database FROM activitydataset")

    return {x[0] for x in result}



primary_types = {
    "key": object,
    "id": "int64",
    "code": str,
    "database": "category",
    "location": "category",
    "name": str,
    "product": object,
    "type": "category",
}
secondary_types = {
    "synonyms": object,
    "unit": "category",
    "CAS number": "category",
    "categories": object,
    "processor": object,
    "allocation": "category",
    "allocation_factor": float,
    "properties": object,
}
all_types = {**primary_types, **secondary_types}

primary = list(primary_types.keys())
secondary = list(secondary_types.keys())
all = primary + secondary



from logging import getLogger

import bw2data as bd
import bw_functional as bf

from activity_browser import signals, bwutils, application
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.ui.icons import qicons

log = getLogger(__name__)


class ActivityOpen(ABAction):
    """
    ABAction to open one or more activities.

    This action processes a list of activities, validates their types, and opens
    their details in the application's central widget. Unsupported activity types
    are logged as warnings.

    Attributes:
        icon (QIcon): The icon representing this action.
        text (str): The display text for this action.
    """

    icon = qicons.right
    text = "Open activity / activities"

    @staticmethod
    @exception_dialogs
    def run(activities: list[tuple | int | bd.Node]):
        """
        Execute the action to open activities.

        This method refreshes the provided activities, validates their types, and
        opens their details in the "Activity Details" group of the central widget.

        Args:
            activities (list[tuple | int | bd.Node]): A list of activities to process.

        Logs:
            Warning: If an activity type is not supported.
        """
        from activity_browser.layouts import pages

        # Refresh the activity nodes to ensure they are up-to-date
        activities = [bwutils.refresh_node(activity) for activity in activities]
        processes = [bwutils.refresh_node(function["processor"]) for function in activities if isinstance(function, bf.Product)]
        activities = list(set(activities + processes))

        for act in activities:
            # Check if the activity type is supported
            if not bwutils.is_node_process(act):
                log.warning(f"Can't open activity {act.key} - opening type: `{act.get('type')}` not supported")
                continue

            # Create a details page for the activity
            page = pages.ActivityDetailsPage(act)
            central = application.main_window.centralWidget()

            # Add the details page to the "Activity Details" group in the central widget
            central.addToGroup("Activity Details", page)



import bw2data
import activity_browser.mod.bw2analyzer as bw2analyzer
import activity_browser.mod.bw2io as bw2io
import activity_browser.mod.ecoinvent_interface as ecoinvent_interface
import activity_browser.mod.pyprind as pyprind
import activity_browser.mod.tqdm as tqdm
import activity_browser.mod.peewee as peewee



import subprocess
import sqlite3
import sys
import pickle
from logging import getLogger
from typing import Literal

import pandas as pd
import bw2data as bd
from bw2data.backends import sqlite3_lci_db

from qtpy import QtCore

from activity_browser import signals, application
from activity_browser.ui.core import threading

from .metadata import MetaDataStore
from .fields import secondary_types, primary, secondary

log = getLogger(__name__)


class MDSLoader(QtCore.QObject):
    primary_status: Literal["idle", "loading", "done"] = "idle"
    secondary_status: Literal["idle", "loading", "done"] = "idle"

    def __init__(self, mds: MetaDataStore):
        super().__init__(mds)

        self.mds = mds
        self.connect_signals()

    def connect_signals(self):
        signals.project.changed.connect(self.on_project_changed)

    def on_project_changed(self):
        self.load_project()

    def load_project(self):
        # set statuses
        self.primary_status = "loading"
        self.secondary_status = "loading"

        # start loading threads
        thread = SecondaryLoadThread(self)
        thread.setObjectName("SecondaryLoadThread-MDSLoader")
        thread.done.connect(self.secondary_load_project)
        thread.start(databases=list(bd.databases), sqlite_db=str(sqlite3_lci_db._filepath))

        # load primary metadata in the main thread
        self.primary_load_project()

    def primary_load_project(self):
        with sqlite3.connect(sqlite3_lci_db._filepath) as con:
            fields = ', '.join(primary[1:])  # Exclude 'key' as it's constructed
            primary_df = pd.read_sql(f"SELECT {fields} FROM activitydataset", con)

        primary_df["key"] = list(zip(primary_df["database"], primary_df["code"]))
        primary_df.index = pd.MultiIndex.from_tuples(primary_df["key"], names=["database", "code"])

        log.debug(f"Primary metadata loaded with {len(primary_df)} rows")
        self.mds.dataframe = primary_df

        for idx in primary_df.index:
            self.mds.register_mutation(idx, "add")
        
        self.primary_status = "done"

    def secondary_load_project(self, secondary_df: pd.DataFrame, sqlite_db: str):
        if sqlite_db != str(sqlite3_lci_db._filepath):
            return

        assert all(secondary_df.index.isin(self.mds.dataframe.index))
        log.debug(f"Secondary metadata loaded with {len(secondary_df)} rows")
        self.mds.dataframe = pd.concat([self.mds.dataframe[primary], secondary_df], axis=1)

        for idx in secondary_df.index:
            self.mds.register_mutation(idx, "update")
        
        self.secondary_status = "done"

    def load_database(self, database_name: str):
        # start loading threads
        thread = SecondaryLoadThread(self)
        thread.done.connect(self.secondary_load_database)
        thread.start(databases=[database_name], sqlite_db=str(sqlite3_lci_db._filepath))

        # load primary metadata in the main thread
        self.primary_load_database(database_name)

    def primary_load_database(self, database_name: str):
        with sqlite3.connect(sqlite3_lci_db._filepath) as con:
            fields = ', '.join(primary[1:])  # Exclude 'key' as it's constructed
            primary_df = pd.read_sql(f"SELECT {fields} FROM activitydataset WHERE database = '{database_name}'", con)

        primary_df["key"] = list(zip(primary_df["database"], primary_df["code"]))
        primary_df.index = pd.MultiIndex.from_tuples(primary_df["key"], names=["database", "code"])

        log.debug(f"Primary metadata loaded with {len(primary_df)} rows")
        self.mds.dataframe = pd.concat([self.mds.dataframe, primary_df])

        for idx in primary_df.index:
            self.mds.register_mutation(idx, "add")

    def secondary_load_database(self, secondary_df: pd.DataFrame, sqlite_db: str):
        if secondary_df.empty or sqlite_db != str(sqlite3_lci_db._filepath):
            return

        database = secondary_df.index[0][0]
        indices = self.mds.dataframe.loc[[database]].index

        if not all(secondary_df.index.isin(indices)):
            log.debug("Secondary database metadata dropping rows")
            secondary_df = secondary_df[secondary_df.index.isin(indices)]

        log.debug(f"Secondary metadata loaded with {len(secondary_df)} rows")

        self._fix_categories(secondary_df)
        self.mds.dataframe.update(secondary_df)

        for idx in secondary_df.index:
            self.mds.register_mutation(idx, "update")

    # utility functions
    def _fix_categories(self, df: pd.DataFrame):
        category_columns = [k for k, v in secondary_types.items() if v == "category"]

        for col in category_columns:
            categories = df[col].dropna().unique()
            categories = [c for c in categories if c not in self.mds.dataframe[col].cat.categories]

            # add new category to column
            self.mds.dataframe[col] = self.mds.dataframe[col].cat.add_categories(categories)


class SecondaryLoadThread(threading.ABThread):
    done: QtCore.SignalInstance = QtCore.Signal(pd.DataFrame, str)

    def run_safely(self, databases: list[str], sqlite_db: str):
        processes = [self.open_load_process(db, sqlite_db) for db in databases]

        full_df = pd.DataFrame()
        for proc in processes:
            stdout_data, stderr_data = proc.communicate()
            if proc.returncode != 0:
                log.error(f"Error loading metadata: {stderr_data.decode()}")
                continue
            df = pickle.loads(stdout_data)
            if df.empty:
                continue

            full_df = pd.concat([full_df, df])

        self.done.emit(full_df, sqlite_db)

    def open_load_process(self, database_name: str, sqlite_db: str) -> subprocess.Popen:
        import activity_browser.bwutils.metadata._sub_loader as sl

        return subprocess.Popen(
            [sys.executable, sl.__file__, str(sqlite_db), database_name] + secondary,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )




from qtpy import QtWidgets

from activity_browser import signals, actions
from activity_browser.ui import widgets, icons

from .scenario_section import ScenarioSection
from .functional_unit_section import FunctionalUnitSection
from .impact_category_section import ImpactCategorySection


class CalculationSetupPage(QtWidgets.QWidget):

    def __init__(self, cs_name: str, parent=None):
        super().__init__(parent)
        self.setObjectName(cs_name)

        self.calculation_setup_name = cs_name

        self.type_dropdown = QtWidgets.QComboBox()
        self.type_dropdown.addItems(["Standard", "Scenario"])

        self.run_button = QtWidgets.QPushButton("Run", self)
        self.run_button.setIcon(icons.qicons.forward)
        self.run_button.setStyleSheet("background-color: #57965C;")

        self.functional_unit_section = FunctionalUnitSection(cs_name, self)
        self.impact_category_section = ImpactCategorySection(cs_name, self)
        self.scenario_section = ScenarioSection(self)
        self.scenario_section.hide()

        # Build the layout of the widget
        self.build_layout()
        self.sync()
        self.connect_signals()

    def build_layout(self):
        """
        Builds the layout of the widget.
        """
        layout = QtWidgets.QVBoxLayout()
        layout.setContentsMargins(0, 3, 0, 0)

        top_layout = QtWidgets.QHBoxLayout()
        top_layout.setContentsMargins(0, 0, 10, 0)
        top_layout.addWidget(widgets.ABLabel.demiBold("    Functional Units:", self))
        top_layout.addStretch()
        top_layout.addWidget(self.type_dropdown)
        top_layout.addWidget(self.run_button)

        # Add fu label and view to the layout
        layout.addLayout(top_layout)
        layout.addWidget(self.functional_unit_section)

        # Add ic label and view to the layout
        layout.addWidget(widgets.ABLabel.demiBold("    Impact Categories:", self))
        layout.addWidget(self.impact_category_section)

        # Add scenario label and view to the layout

        layout.addWidget(self.scenario_section)

        # Set the layout for the widget
        self.setLayout(layout)

    def connect_signals(self):
        signals.project.changed.connect(self.sync)
        signals.meta.calculation_setups_changed.connect(self.sync)

        self.type_dropdown.currentTextChanged.connect(self.type_switch)
        self.run_button.released.connect(self.run_calculation)

    def sync(self) -> None:
        self.functional_unit_section.sync()
        self.impact_category_section.sync()

    def type_switch(self, calculation_type: str):
        if calculation_type == "Standard":
            self.scenario_section.hide()
        elif calculation_type == "Scenario":
            self.scenario_section.show()
        else:
            raise ValueError(f"Unknown calculation type: {calculation_type}")

    def run_calculation(self):
        if self.type_dropdown.currentText() == "Standard":
            actions.CSCalculate.run(self.calculation_setup_name)
        elif self.type_dropdown.currentText() == "Scenario":
            scenario_data = self.scenario_section.scenario_dataframe()
            actions.CSCalculate.run(self.calculation_setup_name, scenario_data)




import json
from tarfile import open as tar_open, TarFile, TarError
from logging import getLogger

from qtpy import QtWidgets, QtCore
from bw2io import restore_project_directory

from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.mod import bw2data as bd
from activity_browser.ui import icons, widgets

log = getLogger(__name__)


class ProjectLocalImportWindow(QtWidgets.QDialog):

    MAX_PROJECT_NAME_JSON_SIZE = 1024
    PROJECT_FILE = ".project-name.json"

    def __init__(self):
        super().__init__()
        self.setWindowTitle("Import project from file")
        self.setSizePolicy(
            QtWidgets.QSizePolicy(
                QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding
            )
        )

        layout = QtWidgets.QVBoxLayout()
        dialog_spacing = 8

        file_chooser_layout = QtWidgets.QHBoxLayout()
        file_chooser_layout.setAlignment(QtCore.Qt.AlignLeft)
        tarball_label = QtWidgets.QLabel("Project file:")
        self._selected_file_edit = QtWidgets.QLineEdit()
        self._selected_file_edit.setMinimumWidth(300)
        self._selected_file_edit.textChanged.connect(self._load_project_name)
        self._browse_button = QtWidgets.QPushButton("Browse")
        self._browse_button.clicked.connect(self._handle_browse_clicked)
        file_chooser_layout.addWidget(tarball_label)
        file_chooser_layout.addWidget(self._selected_file_edit)
        file_chooser_layout.addWidget(self._browse_button)

        layout.addLayout(file_chooser_layout)

        layout.addSpacing(dialog_spacing)

        project_name_layout = QtWidgets.QHBoxLayout()
        project_name_layout.setAlignment(QtCore.Qt.AlignLeft)
        project_name_layout.addWidget(widgets.ABLabel.demiBold("Project name:"))
        self.project_name = QtWidgets.QLineEdit()
        self.project_name.setText("")
        self.project_name.textChanged.connect(self._handle_project_name_changed)
        project_name_layout.addWidget(self.project_name)
        layout.addLayout(project_name_layout)

        self._overwrite_checkbox = QtWidgets.QCheckBox("Overwrite existing project")
        self._overwrite_checkbox.clicked.connect(self._handle_overwrite_clicked)
        layout.addWidget(self._overwrite_checkbox)

        self._activate_project_checkbox = QtWidgets.QCheckBox("Activate project after import")
        self._activate_project_checkbox.setChecked(True)
        layout.addWidget(self._activate_project_checkbox)

        import_button_layout = QtWidgets.QHBoxLayout()
        self.import_button = QtWidgets.QPushButton("Create project")
        import_button_layout.addWidget(self.import_button)
        self.import_button.clicked.connect(self._import_project)
        layout.addLayout(import_button_layout)
        self._message_label = QtWidgets.QLabel()

        layout.addWidget(self._message_label)

        self.setLayout(layout)
        self._last_url = ""
        self._loaded_project_name = ""
        self._reset_dialog()
        self._message_label.setText("Select a project file")

    def _reset_dialog(self):
        self.project_name.setEnabled(False)
        self.project_name.setPlaceholderText("")
        self._overwrite_checkbox.setEnabled(False)
        self._overwrite_checkbox.setChecked(False)
        self._activate_project_checkbox.setEnabled(False)
        self.import_button.setEnabled(False)
        self._message_label.setText("")

    def _enable_ui(self):
        self.project_name.setEnabled(True)
        self.project_name.setPlaceholderText("")
        self._activate_project_checkbox.setEnabled(True)
        self._message_label.setText("")

    def _handle_browse_clicked(self):
        """Open a system file dialog and allow the user to select a file"""
        file = QtWidgets.QFileDialog().getOpenFileName(
            self,
            "Select archive file",
            filter = "Tar GZ (*.tar.gz)"
        )[0]
        # The returned value is None on Cancel
        if file:
            self._selected_file_edit.setText(QtCore.QDir.toNativeSeparators(file))
            self._load_project_name()

    def _decode_project_name(self, tar: TarFile):
        """
        Get the list of files from the TarFile, and decode the name
        from the .project-name.json.

        Updates the UI with error messages if it fails.
        """
        # all files in the archive
        name_list = tar.getnames()
        # list of files, where the path contains ".project-name.json"
        project_name_files = [name for name in name_list if self.PROJECT_FILE in name]
        if len(project_name_files) == 0:
            self._message_label.setText(
                f"No '{self.PROJECT_FILE}' file found in project file"
            )
            return
        if len(project_name_files) > 1:
            self._message_label.setText(
                f"More than one '{self.PROJECT_FILE}' file found in project file"
            )
            return
        # choose the first one, we expect to have only one
        project_name_file = project_name_files[0]
        # get TarInfo for it
        tar_info_project_name_file = tar.getmember(project_name_file)
        # prevent too big files from being extracted
        if tar_info_project_name_file.size > self.MAX_PROJECT_NAME_JSON_SIZE:
            self._message_label.setText(
                f"Size of '{self.PROJECT_FILE}' file is too "
                f"big: {tar_info_project_name_file.size}"
            )
            return
        # get extracter BufferedReader
        if extracter := tar.extractfile(project_name_file):
            try:
                # JSON should have a single string value with the key "name"
                project_name = json.loads(extracter.read())["name"]
            except:
                self._message_label.setText(
                    "Failed to decode project name"
                )
                return
            if project_name == "":
                self._message_label.setText("Decoded project name is empty")
                return
            self._enable_ui()
            self._loaded_project_name = project_name
            self.project_name.setPlaceholderText(self._loaded_project_name)
            self._check_project_already_exists()

    def _load_project_name(self):
        """Exception handling for the project name decoding."""
        try:
            self._reset_dialog()
            archive = self._selected_file_edit.text()
            tar = tar_open(archive, "r:gz")
        except FileNotFoundError:
            self._message_label.setText("Project file not found")
        except TarError:
            self._message_label.setText("Error opening project file")
        except (ValueError, OSError):
            self._message_label.setText("Select a project file")
        else:
            try:
                with tar:
                    self._decode_project_name(tar)
            except TarError:
                self._message_label.setText("Error opening project file")

    def _selected_project_name(self) -> str:
        """The name of the project as decoded from the tarball"""
        return self._loaded_project_name

    def _project_name(self) -> str:
        """Return the user typed project name or, if empty, the loaded one."""
        if self.project_name.text() == "":
            return self._selected_project_name()
        return self.project_name.text()

    def _handle_project_name_changed(self):
        """Trigger duplicate project name check"""
        self._check_project_already_exists()

    def _unique_project_update(self):
        """
        Update the UI when the entered project name is unique.
        """
        self.import_button.setEnabled(True)
        self._message_label.setText("")

    def _duplicate_project_checkbox_update(self):
        """
        Update the UI when the overwrite checkbox state changes and the
        project name is not unique.

        Use the actual state of the checkbox, because it is not
        called only from the checkbox click event.
        """
        if self._overwrite_checkbox.isChecked():
            self.import_button.setEnabled(True)
            self._message_label.setText("")
        else:
            self.import_button.setEnabled(False)
            self._message_label.setText("Project name already exists")

    def _handle_overwrite_clicked(self):
        self._duplicate_project_checkbox_update()

    def _check_project_already_exists(self):
        """
        Update the overwrite checkbox and import button based on the project name.

        If the project already exists, it can only be imported with the
        overwrite flag set. To make sure the user does not import it accidentaly,
        the flag is reset every time the selected project or the project name changes.
        """
        self._overwrite_checkbox.setChecked(False)
        if self._project_name() in bd.projects:
            self._overwrite_checkbox.setEnabled(True)
            self._duplicate_project_checkbox_update()
        else:
            self._overwrite_checkbox.setEnabled(False)
            self._unique_project_update()

    def _import_project(self):
        """
        Import the selected project with the new name.
        It is checked with the UI flow, that there is a tarball loaded,
        and a unique name provided or the overwrite flag is set.
        """
        original_name = self._selected_project_name()
        new_name = self._project_name()
        if original_name and new_name:
            log.info(f"Importing project with name {new_name} "
                     f"(original name {original_name})")
            self.import_button.setText("Creating project...")
            self.import_button.setEnabled(False)
            self.repaint()
            self.setCursor(QtCore.Qt.WaitCursor)

            restore_project_directory(
                self._selected_file_edit.text(),
                new_name,
                overwrite_existing=self._overwrite_checkbox.isChecked()
            )
            if self._activate_project_checkbox.isChecked():
                bd.projects.set_current(new_name, update=False)
            self.setCursor(QtCore.Qt.ArrowCursor)
            self.accept()
        else:
            log.error(
                f"Project name ({new_name}) or "
                f"import name ({original_name}) is not valid."
            )


class ProjectLocalImport(ABAction):
    """
    ABAction to download a project file from a remote server.
    Allows for customization of server URL, created project name, and whether or not to overwrite existing projects.
    """

    icon = icons.qicons.import_db
    text = "Import local project"
    tool_tip = "Import a project file from a remote server"

    @staticmethod
    @exception_dialogs
    def run():
        window = ProjectLocalImportWindow()
        window.adjustSize()
        window.exec_()



import os

from qtpy import QtWebEngineWidgets, QtWidgets, QtCore, QtGui, QtWebChannel

from activity_browser import actions, signals
from activity_browser.static import startscreen
from activity_browser.bwutils import projects_by_last_opened


class WelcomePage(QtWidgets.QWidget):
    html_file = os.path.join(startscreen.__path__[0], "welcome.html")

    def __init__(self, parent=None):
        super().__init__(parent)
        self.view = QtWebEngineWidgets.QWebEngineView()
        self.page = WelcomeWebPage()
        self.channel = QtWebChannel.QWebChannel(self)
        self.bridge = Bridge(self)
        self.channel.registerObject("bridge", self.bridge)

        self.url = QtCore.QUrl.fromLocalFile(self.html_file)
        self.page.setWebChannel(self.channel)
        self.page.load(self.url)

        # associate page with view
        self.view.setPage(self.page)

        # set layout
        self.vl = QtWidgets.QVBoxLayout()
        self.vl.addWidget(self.view)
        self.setLayout(self.vl)

        self.bridge.ready.connect(self.update_welcome)
        signals.project.changed.connect(lambda: self.page.load(self.url))

    def update_welcome(self):
        projects = projects_by_last_opened()
        projects = projects[1:5] if len(projects) > 5 else projects[1:]
        project_names = [p.name for p in projects]
        self.bridge.update.emit(project_names)


class Bridge(QtCore.QObject):
    """
    A bridge for communication between Python and JavaScript.

    Attributes:
        update_graph (SignalInstance): A signal to update the graph.
        ready (SignalInstance): A signal indicating that the bridge is ready.
    """
    update: QtCore.SignalInstance = QtCore.Signal(list)
    ready: QtCore.SignalInstance = QtCore.Signal()

    @QtCore.Slot()
    def is_ready(self):
        """
        Emits the ready signal.
        """
        self.ready.emit()

    @QtCore.Slot(str)
    def open_project(self, project_name):
        """
        Emits the ready signal.
        """
        actions.ProjectSwitch.run(project_name)

class WelcomeWebPage(QtWebEngineWidgets.QWebEnginePage):
    def acceptNavigationRequest(self, qurl, navtype, mainframe):
        # print("Navigation Request intercepted:", qurl)
        if qurl.isLocalFile():  # open in Activity Browser QWebEngineView
            return True
        else:  # delegate link to default browser
            QtGui.QDesktopServices.openUrl(qurl)
            return False



# -*- coding: utf-8 -*-
from pathlib import Path

from qtpy.QtCore import Qt, QSize
from qtpy.QtGui import QIcon, QPixmap

PACKAGE_DIR = Path(__file__).resolve().parents[1]


def create_path(folder: str, filename: str) -> str:
    """Builds a path to the image file."""
    return str(PACKAGE_DIR.joinpath("static", "icons", folder, filename))


def empty_icon(size: QSize = QSize(32, 32)) -> QIcon:
    pixmap = QPixmap(size)
    pixmap.fill(Qt.transparent)  # Make the pixmap transparent
    return QIcon(pixmap)


# CURRENTLY UNUSED ICONS

# Modular LCA (keep until this is reintegrated)
# add_db = create_path('metaprocess', 'add_database.png')
# close_db = create_path('metaprocess', 'close_database.png')
# cut = create_path('metaprocess', 'cut.png')
# debug = create_path('main', 'ladybird.png')
# duplicate = create_path('metaprocess', 'duplicate.png')
# graph_lmp = create_path('metaprocess', 'graph_linkedmetaprocess.png')
# graph_mp = create_path('metaprocess', 'graph_metaprocess.png')
# load_db = create_path('metaprocess', 'open_database.png')
# metaprocess = create_path('metaprocess', 'metaprocess.png')
# new = create_path('metaprocess', 'new_metaprocess.png')
# save_db = create_path('metaprocess', 'save_database.png')
# save_mp = create_path('metaprocess', 'save_metaprocess.png')

# key = create_path('main', 'key.png')
# search = create_path('main', 'search.png')
# switch = create_path('main', 'switch-state.png')


class Icons(object):
    # Icons from href="https://www.flaticon.com/

    # MAIN
    ab = create_path("main", "activitybrowser.png")

    # arrows
    right = create_path("main", "right.png")
    left = create_path("main", "left.png")
    forward = create_path("main", "forward.png")
    backward = create_path("main", "backward.png")

    # Simple actions
    delete = create_path("context", "delete.png")
    clear = create_path("context", "clear.png")
    copy = create_path("context", "copy.png")
    add = create_path("context", "add.png")
    edit = create_path("main", "edit.png")
    calculate = create_path("main", "calculate.png")
    question = create_path("context", "question.png")
    search = create_path("main", "search.png")
    filter = create_path("main", "filter.png")
    filter_outline = create_path("main", "filter_outline.png")

    # database
    import_db = create_path("main", "import_database.png")
    duplicate_database = create_path("main", "duplicate_database.png")

    # activity
    duplicate_activity = create_path("main", "duplicate_activity.png")
    duplicate_to_other_database = create_path("main", "import_database.png")
    parameterized = create_path("main", "parameterized.png")

    # windows
    graph_explorer = create_path("main", "graph_explorer.png")
    issue = create_path("main", "idea.png")
    settings = create_path("main", "settings.png")
    history = create_path("main", "history.png")
    welcome = create_path("main", "welcome.png")
    main_window = create_path("main", "home.png")

    # plugins
    plugin = create_path("main", "plugin.png")

    # nodes
    process = create_path("nodes", "process.png")
    product = create_path("nodes", "product.png")
    waste = create_path("nodes", "waste.png")
    processproduct = create_path("nodes", "processproduct.png")
    biosphere = create_path("nodes", "biosphere.png")
    readonly_process = create_path("nodes", "read-only-process.png")

    # other
    superstructure = create_path("main", "superstructure.png")
    copy_to_clipboard = create_path("main", "copy_to_clipboard.png")
    warning = create_path("context", "warning.png")
    critical = create_path("context", "critical.png")
    locked = create_path("main", "locked.png")
    unlocked = create_path("main", "unlocked.png")


class QIcons(Icons):
    """Using the Icons class, returns the same attributes, but as QIcon type"""
    empty = empty_icon()

    def __getattribute__(self, item):
        return QIcon(Icons.__getattribute__(self, item))


icons = Icons()
qicons = QIcons()



from typing import TYPE_CHECKING
from qtpy import QtWidgets

if TYPE_CHECKING:
    from activity_browser.ui.widgets import ABWizardPage


class ABWizard(QtWidgets.QWizard):
    pages = []

    def __init__(self, *args, title: str = None, context: dict = None, **kwargs):
        super().__init__(*args, **kwargs)

        self.setWizardStyle(QtWidgets.QWizard.WizardStyle.ModernStyle)

        if title:
            self.setWindowTitle(title)

        for page in self.pages:
            self.addPage(page(self))

        self.context = context or {}

    def page(self, page_id: int) -> "ABWizardPage":
        return super().page(page_id)

    def initializePage(self, page_id):

        # finalize the previous page if it exists
        if self.visitedIds():
            previous_page_id = self.visitedIds()[-1]
            previous_page = self.page(previous_page_id)
            previous_page.finalize(self.context)

        # initialize the next page
        page = self.page(page_id)
        page.initializePage(self.context)



from logging import getLogger

import bw2data as bd
from bw2data.parameters import ParameterBase, parameters, ActivityParameter, Group, GroupDependency
from peewee import DoesNotExist

from activity_browser.ui.icons import qicons
from activity_browser.bwutils import refresh_parameter, Parameter
from activity_browser.actions.base import ABAction, exception_dialogs

from .parameter_rename import ParameterRename

log = getLogger(__name__)


class ParameterModify(ABAction):
    """
    ABAction to delete an existing parameter.
    """

    icon = qicons.edit
    text = "Modify Parameter"

    @staticmethod
    @exception_dialogs
    def run(parameter: tuple | Parameter | ParameterBase, field: str, value: any):
        parameter = refresh_parameter(parameter)
        param_model = parameter.to_peewee_model()

        if field == "data":
            param_model.data.update(value)
        elif field == "name":
            return ParameterRename.run(parameter, value)
        elif field in dir(param_model):
            setattr(param_model, field, value)
        else:
            param_model.data.update({field: value})

        param_model.save()

        if field in ("amount", "formula"):
            ParameterModify.fix_broken_groups()
            parameters.recalculate()

    @staticmethod
    def fix_broken_groups():
        groups = Group.select().execute()
        for group in groups:
            if group.name == "project" or group.name in bd.databases:
                continue
            try:
                ActivityParameter._static_dependencies(group.name)
            except DoesNotExist:
                log.warning(f"Removing broken parameter group {group.name}")
                GroupDependency.get(GroupDependency.group == group.name).delete_instance()
                group.delete_instance()
